# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

# pylint: disable=wildcard-import, unused-wildcard-import
import datetime, os, platform, struct, time, zipfile
from serial import SerialException
from .app import AppInfo
from .LK_Device import *
from .i18n import __, c__
from .IniSettings import IniSettings

class GbxDevice(LK_Device):
	DEVICE_NAME = "Joey Jr"
	DEVICE_MIN_FW = 1
	DEVICE_MAX_FW = 12
	DEVICE_LATEST_FW_TS = 1780508702
	PCB_VERSIONS = { -1:"", 0x01:"V2", 0x81:"V2", 0x02:"V2C", 0x82:"V2C", 0x03:"V2CC", 0x83:"V2CC/V2++" }
	DEVICE_LABEL_LONG = "Joey Jr"
	DEVICE_LABEL_SHORT = "Joey Jr"
	FWUPDATE_ACTION = "fwupdate-joeyjr"
	CLI_UPDATER_METHOD = "UpdateFirmwareJoeyJr"
	DEVICE_SUPPORT_MESSAGE = "For help with your Joey Jr, please visit the BennVenn Discord:\nhttps://discord.gg/F5ckxM2"
	MAX_BUFFER_READ = 0x1000
	MAX_BUFFER_WRITE = 0x800

	def __init__(self):
		pass

	def Initialize(self, flashcarts, port=None, max_baud=2000000):
		if self.IsConnected(): self.DEVICE.close()
		conn_msg = []
		ports = []
		if port is not None:
			ports = [ port ]
		else:
			comports = serial.tools.list_ports.comports()
			for i in range(0, len(comports)):
				if comports[i].vid == 0x483 and comports[i].pid == 0x5740:
					ports.append(comports[i].device)
			if len(ports) == 0: return False

		for i in range(0, len(ports)):
			if self.TryConnect(ports[i], max_baud):
				self.BAUDRATE = max_baud
				dev = serial.Serial(ports[i], self.BAUDRATE, timeout=0.1)
				self.DEVICE = dev
			else:
				continue

			if self.FW is None or self.FW == {}: continue

			dprint(f"Found a {self.DEVICE_NAME}")
			dprint("Firmware information:", self.FW)
			# dprint("Baud rate:", self.BAUDRATE)

			if self.DEVICE is None or not self.IsConnected():
				self.DEVICE = None
				if self.FW is not None:
					conn_msg.append([0, __("Couldn’t communicate with the {device_name} on port {port}. Please disconnect and reconnect the device, then try again.", device_name=self.DEVICE_NAME, port=ports[i])])
				continue
			elif self.FW is None:
				dev.close()
				self.DEVICE = None
				continue
			elif self.FW["cfw_id"] == "G": # Not a CFW by Lesserkuma
				dprint("Device runs the JoeyGUI firmware")
			elif self.FW["pcb_ver"] not in self.PCB_VERSIONS.keys() or "cfw_id" not in self.FW or self.FW["cfw_id"] != 'L' or self.FW["fw_ver"] < self.DEVICE_MIN_FW: # Not a CFW by Lesserkuma
				dprint("Incompatible firmware:", self.FW)
				dev.close()
				self.DEVICE = None
				continue
			elif self.FW["fw_ts"] > self.DEVICE_LATEST_FW_TS:
				conn_msg.append([1, __("Note: The {device_name} on port {port} is running a firmware version that is newer than what this version of FlashGBX was developed to work with, so errors may occur.", device_name=self.DEVICE_NAME, port=ports[i])])

			if (self.FW["pcb_ver"] & 0x7F) == 1:
				conn_msg.append([2, __("Warning: Your {device_name} does not support the software-controlled voltage setting feature, so please be extra careful and always set the switch to the correct voltage before inserting a cartridge.", device_name=self.DEVICE_NAME)])

			self.PORT = ports[i]
			self.DEVICE.timeout = self.DEVICE_TIMEOUT

			# Load Flash Cartridge Handlers
			self.UpdateFlashCarts(flashcarts)

			# Stop after first found device
			break

		return conn_msg

	def LoadFirmwareVersion(self):
		dprint("Querying firmware version")
		try:
			self.DEVICE.timeout = 0.075
			self.DEVICE.reset_input_buffer()
			self.DEVICE.reset_output_buffer()

			self._write(bytearray(b'\x55\xAA'))
			time.sleep(0.01)
			device_id = self.DEVICE.read(self.DEVICE.in_waiting)

			if b"Joey" not in device_id:
				dprint("Not a Joey Jr")
				self.FW = None
				return False

			if b"FW L" not in device_id:
				dprint("Not running LK firmware")
				if b"GUI" in device_id or b"FW vG" in device_id:
					pcb_name = device_id[1:].decode("UTF-8", "ignore")
					self.FW = {
						"pcb_ver":-1,
						"pcb_name":f"{pcb_name}",
						"cfw_id":"G",
						"fw_ver":0,
						"fw_ts":0,
						"fw_dt":"JoeyGUI",
						"cart_power_ctrl":False,
						"bootloader_reset":True
					}
					return True
				return False

			if device_id[0] == 0:
				self._write(bytearray(b'LK')) # Enable LK firmware
				if self.DEVICE.read(1) != b'\xFF':
					dprint("LK firmware was not enabled successfully")
					self.FW = None
					return False

			self._write(self.DEVICE_CMD["QUERY_FW_INFO"])
			size = self.DEVICE.read(1)
			self.DEVICE.timeout = self.DEVICE_TIMEOUT
			if len(size) == 0:
				dprint("No response")
				self.FW = None
				return False
			size = struct.unpack("B", size)[0]
			if size != 8:
				print(size)
				return False
			data = self._read(size)
			info = data[:8]
			keys = ["cfw_id", "fw_ver", "pcb_ver", "fw_ts"]
			values = struct.unpack(">cHBI", bytearray(info))
			self.FW = dict(zip(keys, values))
			self.FW["cfw_id"] = self.FW["cfw_id"].decode('ascii')
			self.FW["fw_dt"] = datetime.datetime.fromtimestamp(self.FW["fw_ts"]).astimezone().replace(microsecond=0).isoformat()
			self.FW["ofw_ver"] = None
			self.FW["pcb_name"] = None
			self.FW["cart_power_ctrl"] = False
			self.FW["bootloader_reset"] = False
			if self.FW["cfw_id"] == "L" and self.FW["fw_ver"] >= 12:
				size = self._read(1)
				name = self._read(size)
				if len(name) > 0:
					try:
						self.FW["pcb_name"] = name.decode("UTF-8").replace("\x00", "").strip()
					except:
						self.FW["pcb_name"] = "Unnamed Device"
					self.DEVICE_NAME = self.FW["pcb_name"]

				# Cartridge Power Control support
				temp = self._read(1)
				self.FW["cart_power_ctrl"] = True if temp & 1 == 1 else False
				self.FW["cart_presence_switch"] = True if (temp >> 1) & 1 == 1 else False
				self.FW["cart_mode_switch"] = True if (temp >> 2) & 1 == 1 else False

				# Reset to bootloader support
				self.FW["bootloader_reset"] = True if self._read(1) == 1 else False
			return True

		except Exception as e:
			dprint("Disconnecting due to an error", e, sep="\n")
			try:
				if self.DEVICE.isOpen():
					self.DEVICE.reset_input_buffer()
					self.DEVICE.reset_output_buffer()
					self.DEVICE.close()
				self.DEVICE = None
			except:
				pass
			return False

	def ChangeBaudRate(self, _):
		dprint("Baudrate change is not supported.")

	def GetFirmwareVersion(self, more=False):
		if self.FW["pcb_ver"] == -1:
			return "JoeyGUI"

		base = "{:s}{:d}".format(self.FW["cfw_id"], self.FW["fw_ver"])
		if self.FW["pcb_name"] is None:
			base = base + " <" + __("unverified") + ">"
		if more:
			return "{base} ({dt})".format(base=base, dt=self.FW["fw_dt"])
		return base

	def GetFullNameLabel(self):
		if self.FW["pcb_ver"] == -1:
			return self.FW["pcb_name"]
		return super().GetFullNameLabel()

	def GetFullName(self):
		return self.GetName()

	def GetFullNameExtended(self, more=False):
		if self.FW["pcb_ver"] == -1:
			return self.FW["pcb_name"]

		if more:
			return __("{device_name} – Firmware {fw_version} ({timestamp}) on {port}", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), timestamp=self.FW["fw_dt"], port=self.GetPort())
		else:
			return __("{device_name} – Firmware {fw_version} ({port})", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), port=self.GetPort())

	def CanSetVoltageBySwitch(self):
		return ((self.FW["pcb_ver"] & 0x7F) == 1)

	def CanSetVoltageByCode(self):
		if ((self.FW["pcb_ver"] & 0x7F) == 1): return False
		return True

	def CanSetVoltageByAutoswitch(self):
		return False

	def CanPowerCycleCart(self):
		return self.FW["cart_power_ctrl"]

	def GetSupprtedModes(self):
		return ["DMG", "AGB"]

	def IsSupported3dMemory(self):
		return True

	def IsClkConnected(self):
		return True

	def SupportsFirmwareUpdates(self):
		return True

	def FirmwareUpdateAvailable(self):
		if self.FW["cfw_id"] == "G":
			self.FW_UPDATE_REQ = True
			return True
		if self.FW["fw_ts"] != self.DEVICE_LATEST_FW_TS:
			return True
		self.FW_UPDATE_REQ = False
		return False

	def GetFirmwareUpdaterClass(self):
		try:
			return (None, FirmwareUpdaterWindow)
		except:
			return None

	def ResetLEDs(self):
		pass

	def SupportsBootloaderReset(self):
		return self.FW["bootloader_reset"]

	def BootloaderReset(self):
		if not self.SupportsBootloaderReset(): return False
		dprint("Resetting to bootloader...")
		try:
			self._write(self.DEVICE_CMD["BOOTLOADER_RESET"], wait=True)
			self._write(1)
			self.Close()
			return True
		except Exception as e:
			print(__("Disconnecting..."), e)
			return False

	def SupportsAudioAsWe(self):
		return True

	def Close(self, cartPowerOff=False):
		if self.FW["cfw_id"] == "G":
			self.DEVICE.close()

		if self.IsConnected():
			dprint("Disconnecting from the device")
			try:
				if cartPowerOff and self.CanPowerCycleCart():
					self._set_fw_variable("AUTO_POWEROFF_TIME", 0)
					self._write(self.DEVICE_CMD["CART_PWR_OFF"], wait=self.FW["fw_ver"] >= 12)
				else:
					self._write(self.DEVICE_CMD["SET_VOLTAGE_3_3V"], wait=self.FW["fw_ver"] >= 12)
				self.DEVICE.write(b'KL') # Disable LK firmware
				self.DEVICE.read(1)
				self.DEVICE.close()
			except:
				self.DEVICE = None
			self.MODE = None


class FirmwareUpdater():
	PORT = None
	DEVICE = None

	def __init__(self, app_path=".", port=None):
		self.APP_PATH = app_path
		self.PORT = port

	def CalcChecksum(self, buffer):
		checksum = 0
		for value in buffer[:0xFFFC]:
			checksum += value
		return checksum

	def TryConnect(self, port):
		return True

	def WriteFirmwareMSC(self, path, buffer, fncSetStatus):
		file = path
		path = os.path.dirname(path) + os.sep
		fncSetStatus(text=__("Connecting... This may take a moment."))

		filename = os.path.split(file)[1]
		filepath = os.path.split(file)[0]
		with open(filepath + os.sep + filename, "rb") as f: temp = f.read().decode("UTF-8", "ignore")
		if not temp.startswith("UPDATE"):
			with open(file, "wb") as f:
				temp = bytearray(b"UPDATE")
				temp += bytearray([0] * (256 - len(temp)))
				f.write(temp)
			hp = 30
			while hp > 0:
				if os.path.exists(path + "FIRMWARE.JR"): break
				time.sleep(1)
				hp -= 1
			if hp == 0:
				fncSetStatus(text=__("Couldn’t communicate with the Joey Jr device."))
				return 2

		try:
			with open(filepath + os.sep + filename, "rb") as f: temp = f.read().decode("UTF-8", "ignore")
		except FileNotFoundError as e:
			try:
				if filename == "MODE.TXT":
					with open(filepath + os.sep + "MODE!.TXT", "rb") as f: temp = f.read().decode("UTF-8", "ignore")
				else:
					raise FileNotFoundError from e
			except FileNotFoundError:
				fncSetStatus(text=__("Couldn’t access the {mode_txt} file. Remove cartridge and try again.", mode_txt="MODE.TXT"))
				return 2

		if not temp.startswith("UPDATE"):
			fncSetStatus(text=__("Couldn’t enter {update} mode. Please try again.", update="UPDATE"))
			return 2

		fncSetStatus(text=__("Updating firmware... Do not unplug the device!"), setProgress=0)
		os.unlink(path + "FIRMWARE.JR")
		if os.path.exists(path + "FIRMWARE.JR"):
			fncSetStatus(text=__("Couldn’t write new firmware. Please try again."))
			return 2

		try:
			f = open(path + "FIRMWARE.JR", "wb")
		except OSError:
			fncSetStatus(text=__("Couldn’t write new firmware. Please try again."))
			return 2

		for i in range(0, len(buffer), 64):
			f.write(buffer[i:i+64])
			percent = float(i + 64) / len(buffer) * 100
			fncSetStatus(text=__("Updating firmware... Do not unplug the device!"), setProgress=percent)

		try:
			f.close()
		except OSError:
			pass

		if b"Joey Jr. Firmware" not in buffer:
			hp = 5
			while hp > 0:
				if not os.path.exists(path + "FIRMWARE.JR"): break
				time.sleep(1)
				hp -= 1
			if hp == 0:
				fncSetStatus(text=__("Couldn’t verify. Please try again."))
				return 2

		fncSetStatus(__("Done!"))
		time.sleep(2)

		return True

	def WriteFirmware(self, buffer, fncSetStatus):
		if len(buffer) < 0x10000:
			return 3
		if struct.unpack(">I", buffer[0xFFFC:0x10000])[0] != self.CalcChecksum(buffer): return 3

		# Check for serial mode
		if self.PORT is None:
			ports = []
			comports = serial.tools.list_ports.comports()
			for i in range(0, len(comports)):
				if comports[i].vid == 0x483 and comports[i].pid == 0x5740:
					ports.append(comports[i].device)
			if len(ports) == 0:
				return 4
			port = ports[0]
		else:
			port = self.PORT

		while True:
			fncSetStatus(text=__("Connecting..."))
			try:
				dev = serial.Serial(port, 2000000, timeout=0.2)
			except SerialException as e:
				if "Errno 13" in str(e) and platform.system() == "Linux":
					fncSetStatus(text=__("No permission to use device! See the {readme} file.", readme="README.md"), enableUI=True)
				else:
					fncSetStatus(text=__("Device not accessible."), enableUI=True)
				return 2
			except:
				fncSetStatus(text=__("Unknown error while accessing the device."), enableUI=True)
				return 2
			dev.reset_input_buffer()

			fncSetStatus(__("Identifying device..."))
			dev.write(b'\x55\xAA')
			time.sleep(0.01)
			device_id = dev.read(dev.in_waiting)

			if b"Joey" not in device_id:
				# print("Not a Joey Jr")
				fncSetStatus(__("Joey Jr device not found."))
				return 2

			if b"FW L" in device_id and device_id[-1] != 0:
				fncSetStatus(__("Rebooting device..."))
				dev.write(b'\xF1')
				dev.read(1)
				dev.write(b'\x01')
				dev.close()
				time.sleep(0.5)
				continue
			break

		dev.write(b'\xFE\x01')
		time.sleep(0.01)
		dev.read(1)
		# print("")

		size = len(buffer)
		counter = 0
		while counter < size:
			percent = float(counter)/size*100
			fw_buffer = buffer[counter:counter+64]
			dev.write(fw_buffer)
			time.sleep(0.001)
			try:
				temp = dev.read(dev.in_waiting)
			except:
				temp = False
			if temp is not False:
				# print("Flashing...", hex(i), end="\r")
				pass
			elif counter + 64 < size:
				fncSetStatus(text=__("Error! Bad response at {address}!", address="0x{address:X}".format(address=counter)), setProgress=percent)
				return 2

			counter += 64
			fncSetStatus(text=__("Updating firmware... Do not unplug the device!"), setProgress=percent)

		dev.close()
		time.sleep(1)
		fncSetStatus(__("Done!"), setProgress=100)
		time.sleep(2)
		return 1


try:
	from .pyside import QtCore, QtWidgets, QtGui, QDesktopWidget

	class FirmwareUpdaterWindow(QtWidgets.QDialog):
		APP = None
		DEVICE = None
		FWUPD = None
		DEV_NAME = "Joey Jr"
		FW_VER = ""
		PCB_VER = ""

		def __init__(self, app, app_path, file=None, icon=None, device=None):
			QtWidgets.QDialog.__init__(self, app)
			if icon is not None: self.setWindowIcon(QtGui.QIcon(icon))
			self.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")
			self.setWindowTitle("FlashGBX – " + __("Firmware Updater for {device_name}", device_name="Joey Jr"))
			self.setWindowFlags((self.windowFlags() | QtCore.Qt.MSWindowsFixedSizeDialogHint) & ~QtCore.Qt.WindowContextHelpButtonHint)

			self.APP = app
			if device is not None:
				self.FWUPD = FirmwareUpdater(app_path, device.GetPort())
				self.DEV_NAME = device.GetName()
				self.FW_VER = device.GetFirmwareVersion(more=True)
				self.PCB_VER = device.GetPCBVersion()
				self.DEVICE = device
			else:
				self.APP.QT_APP.processEvents()
				self.FWUPD = FirmwareUpdater(app_path, None)

			self.layout = QtWidgets.QGridLayout()
			self.layout.setContentsMargins(-1, 8, -1, 8)
			self.layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
			self.layout_device = QtWidgets.QVBoxLayout()

			# ↓↓↓ Current Device Information
			self.grpDeviceInfo = QtWidgets.QGroupBox(__("Current Firmware"))
			self.grpDeviceInfo.setMinimumWidth(420)
			self.grpDeviceInfoLayout = QtWidgets.QVBoxLayout()
			self.grpDeviceInfoLayout.setContentsMargins(-1, 3, -1, -1)
			rowDeviceInfo1 = QtWidgets.QHBoxLayout()
			self.lblDeviceName = QtWidgets.QLabel(__("Device:"))
			self.lblDeviceName.setMinimumWidth(120)
			self.lblDeviceNameResult = QtWidgets.QLabel("Joey Jr")
			rowDeviceInfo1.addWidget(self.lblDeviceName)
			rowDeviceInfo1.addWidget(self.lblDeviceNameResult)
			rowDeviceInfo1.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo1)
			rowDeviceInfo3 = QtWidgets.QHBoxLayout()
			self.lblDeviceFWVer = QtWidgets.QLabel(__("Firmware version:"))
			self.lblDeviceFWVer.setMinimumWidth(120)
			self.lblDeviceFWVerResult = QtWidgets.QLabel()
			rowDeviceInfo3.addWidget(self.lblDeviceFWVer)
			rowDeviceInfo3.addWidget(self.lblDeviceFWVerResult)
			rowDeviceInfo3.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo3)
			self.grpDeviceInfo.setLayout(self.grpDeviceInfoLayout)
			self.layout_device.addWidget(self.grpDeviceInfo)
			# ↑↑↑ Current Device Information

			# ↓↓↓ Available Firmware Updates
			file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_JoeyJr.zip")

			try:
				with zipfile.ZipFile(file_name) as zip:
					with zip.open("fw.ini") as f: ini_file = f.read()
					ini_file = ini_file.decode(encoding="utf-8")
					self.INI = IniSettings(ini=ini_file, main_section="Firmware")
					self.FW_LK_VER = __("LK firmware version {version} by {author} (updated on {date})", version=self.INI.GetValue("fw_ver"), author="Lesserkuma", date=datetime.datetime.fromtimestamp(int(self.INI.GetValue("fw_buildts"))).strftime("%x"))
					self.FW_LK_BUILDTS = self.INI.GetValue("fw_buildts")
					self.FW_LK_TEXT = "<ul><li>" + __("For use with the FlashGBX software\nNo support by BennVenn").replace("\n", "</li><li>") + "</li></ul>"
					self.FW_MSC_VER = __("BennVenn Drag’n’Drop firmware version {version} (updated on {date})", version=self.INI.GetValue("fw_msc_ver"), date=datetime.datetime.fromtimestamp(int(self.INI.GetValue("fw_msc_buildts"))).strftime("%x"))
					self.FW_MSC_TEXT = "<ul><li>" + __("For use with the Windows file explorer") + "</li></ul>"
					self.FW_JOEYGUI_VER = __("BennVenn JoeyGUI firmware version {version} (updated on {date})", version=self.INI.GetValue("fw_joeygui_ver"), date=datetime.datetime.fromtimestamp(int(self.INI.GetValue("fw_joeygui_buildts"))).strftime("%x"))
					self.FW_JOEYGUI_TEXT = "<ul><li>" + __("For use with the JoeyGUI software") + "</li></ul>"
			except (FileNotFoundError, zipfile.BadZipFile, KeyError, ValueError):
				QtWidgets.QMessageBox.critical(self, __("Error"), __("The firmware update file is corrupted."))
				self.reject()
				return

			self.grpAvailableFwUpdates = QtWidgets.QGroupBox(__("Firmware Update Options"))
			self.grpAvailableFwUpdates.setMinimumWidth(400)
			self.grpAvailableFwUpdatesLayout = QtWidgets.QVBoxLayout()
			self.grpAvailableFwUpdatesLayout.setContentsMargins(-1, 3, -1, -1)

			self.optFW_LK = QtWidgets.QRadioButton("{:s}".format(self.FW_LK_VER))
			self.lblFW_LK_Info = QtWidgets.QLabel("{:s}".format(self.FW_LK_TEXT))
			self.lblFW_LK_Info.setWordWrap(True)
			self.lblFW_LK_Info.mousePressEvent = lambda x: [ self.optFW_LK.setChecked(True) ]
			self.optFW_MSC = QtWidgets.QRadioButton("{:s}".format(self.FW_MSC_VER))
			self.lblFW_MSC_Info = QtWidgets.QLabel("{:s}".format(self.FW_MSC_TEXT))
			self.lblFW_MSC_Info.setWordWrap(True)
			self.lblFW_MSC_Info.mousePressEvent = lambda x: [ self.optFW_MSC.setChecked(True) ]
			self.optFW_JoeyGUI = QtWidgets.QRadioButton("{:s}".format(self.FW_JOEYGUI_VER))
			self.lblFW_JoeyGUI_Info = QtWidgets.QLabel("{:s}".format(self.FW_JOEYGUI_TEXT))
			self.lblFW_JoeyGUI_Info.setWordWrap(True)
			self.lblFW_JoeyGUI_Info.mousePressEvent = lambda x: [ self.optFW_JoeyGUI.setChecked(True) ]
			self.optExternal = QtWidgets.QRadioButton(__("External FIRMWARE.JR file"))

			self.rowUpdate = QtWidgets.QHBoxLayout()
			self.btnUpdate = QtWidgets.QPushButton(__("Install Firmware Update"))
			self.btnUpdate.setMinimumWidth(200)
			self.btnUpdate.setContentsMargins(20, 20, 20, 20)
			self.connect(self.btnUpdate, QtCore.SIGNAL("clicked()"), lambda: [ self.UpdateFirmware() ])
			self.rowUpdate.addStretch()
			self.rowUpdate.addWidget(self.btnUpdate)
			self.rowUpdate.addStretch()

			self.rowUpdate2 = QtWidgets.QHBoxLayout()
			self.lblUpdateDisclaimer = QtWidgets.QLabel(__("Please note that FlashGBX is not officially supported by BennVenn."))
			self.lblUpdateDisclaimer.setWordWrap(True)
			self.lblUpdateDisclaimer.setAlignment(QtGui.Qt.AlignmentFlag.AlignCenter)
			self.rowUpdate2.addWidget(self.lblUpdateDisclaimer)

			self.grpAvailableFwUpdatesLayout.addWidget(self.optFW_LK)
			self.grpAvailableFwUpdatesLayout.addWidget(self.lblFW_LK_Info)
			self.optFW_LK.setChecked(True)
			self.grpAvailableFwUpdatesLayout.addWidget(self.optFW_MSC)
			self.grpAvailableFwUpdatesLayout.addWidget(self.lblFW_MSC_Info)
			self.grpAvailableFwUpdatesLayout.addWidget(self.optFW_JoeyGUI)
			self.grpAvailableFwUpdatesLayout.addWidget(self.lblFW_JoeyGUI_Info)
			self.grpAvailableFwUpdatesLayout.addWidget(self.optExternal)
			self.grpAvailableFwUpdatesLayout.addSpacing(3)
			self.grpAvailableFwUpdatesLayout.addItem(self.rowUpdate)
			self.grpAvailableFwUpdatesLayout.addItem(self.rowUpdate2)
			self.grpAvailableFwUpdates.setLayout(self.grpAvailableFwUpdatesLayout)
			self.layout_device.addWidget(self.grpAvailableFwUpdates)
			# ↑↑↑ Available Firmware Updates

			self.grpStatus = QtWidgets.QGroupBox("")
			self.grpStatusLayout = QtWidgets.QGridLayout()
			self.prgStatus = QtWidgets.QProgressBar()
			self.prgStatus.setMinimum(0)
			self.prgStatus.setMaximum(1000)
			self.prgStatus.setValue(0)
			self.lblStatus = QtWidgets.QLabel(__("Status: Ready."))

			self.grpStatusLayout.addWidget(self.prgStatus, 1, 0)
			self.grpStatusLayout.addWidget(self.lblStatus, 2, 0)

			self.grpStatus.setLayout(self.grpStatusLayout)
			self.layout_device.addWidget(self.grpStatus)

			self.grpFooterLayout = QtWidgets.QHBoxLayout()
			self.btnClose = QtWidgets.QPushButton(c__("Button (& = Keyboard Shortcut)", "&Close"))
			self.connect(self.btnClose, QtCore.SIGNAL("clicked()"), lambda: [ self.reject() ])
			self.grpFooterLayout.addStretch()
			self.grpFooterLayout.addWidget(self.btnClose)
			self.layout_device.addItem(self.grpFooterLayout)

			self.layout.addLayout(self.layout_device, 0, 0)
			self.setLayout(self.layout)

			self.lblDeviceNameResult.setText(self.DEV_NAME + " " + self.PCB_VER)
			self.lblDeviceFWVerResult.setText(self.FW_VER)

			# if platform.system() == 'Darwin':
			# 	self.optFW_MSC.setVisible(False)
			# 	self.lblFW_MSC_Info.setVisible(False)

		def run(self):
			try:
				self.layout.update()
				self.layout.activate()
				screenGeometry = QDesktopWidget().screenGeometry(self)
				x = (screenGeometry.width() - self.width()) / 2
				y = (screenGeometry.height() - self.height()) / 2
				self.move(x, y)
				self.show()
			except:
				return

		def hideEvent(self, event):
			if self.DEVICE is None:
				self.APP.ConnectDevice()
			self.APP.activateWindow()

		def reject(self):
			if self.CloseDialog():
				super().reject()

		def CloseDialog(self):
			if self.btnClose.isEnabled() is False:
				text = __("<b>Warning:</b> If you close this window while a firmware update is still running, it might leave the device in an unbootable state.") + " " + __("You can still recover it by running the Firmware Updater again later.") + "<br><br>" + __("Are you sure you want to close this window?")
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Warning, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
				msgbox.setDefaultButton(QtWidgets.QMessageBox.No)
				answer = msgbox.exec()
				if answer == QtWidgets.QMessageBox.No: return False
			return True

		def UpdateFirmware(self):
			with zipfile.ZipFile(self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_JoeyJr.zip")) as archive:
				fw = ""
				path = ""
				verified = False
				if self.optFW_LK.isChecked():
					fw = self.FW_LK_VER
					fn = "FIRMWARE_LK.JR"
					with archive.open(fn) as f: fw_data = bytearray(f.read())
					if (b"Joey Jr" in fw_data and b"FW LK" in fw_data): verified = True
				elif self.optFW_MSC.isChecked():
					fw = self.FW_MSC_VER
					fn = "FIRMWARE_MSC.JR"
					with archive.open(fn) as f: fw_data = bytearray(f.read())
					if (b"Joey Jr. Firmware" in fw_data): verified = True
				elif self.optFW_JoeyGUI.isChecked():
					fw = self.FW_JOEYGUI_VER
					fn = "FIRMWARE_JOEYGUI.JR"
					with archive.open(fn) as f: fw_data = bytearray(f.read())
					if (b"Joey Jr" in fw_data and b"FW GUI" in fw_data): verified = True
				else:
					path = self.APP.SETTINGS.value("LastDirFirmwareUpdate")
					path = QtWidgets.QFileDialog.getOpenFileName(self, __("Choose Joey Jr Firmware File"), path, __("Firmware Update ({firmware_jr})", firmware_jr="FIRMWARE.JR"))[0]
					if path == "": return
					if not os.path.basename(path).endswith(".JR"):
						msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=__("The expected filename for a valid firmware file is <b>{filename_pattern}</b>. Please visit {url} for the latest official firmware updates.", firmware_jr="FIRMWARE.JR", url='<a href="https://bennvenn.myshopify.com/products/usb-gb-c-cart-dumper-the-joey-jr">https://bennvenn.myshopify.com/products/usb-gb-c-cart-dumper-the-joey-jr</a>'), standardButtons=QtWidgets.QMessageBox.Ok)
						answer = msgbox.exec()
						return
					elif os.path.exists(os.path.dirname(path) + "DEBUG.TXT"):
						msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=__("The selected file can not be used. Please visit {url} for the latest official firmware updates.", url='<a href="https://bennvenn.myshopify.com/products/usb-gb-c-cart-dumper-the-joey-jr">https://bennvenn.myshopify.com/products/usb-gb-c-cart-dumper-the-joey-jr</a>'), standardButtons=QtWidgets.QMessageBox.Ok)
						answer = msgbox.exec()
						return
					self.APP.SETTINGS.setValue("LastDirFirmwareUpdate", os.path.dirname(path))
					fw = path
					fn = None
					try:
						with open(path, "rb") as f: fw_data = bytearray(f.read())
						index_from = fw_data.index(b"Joey Jr")
						index_to = fw_data[index_from:].index(b"\x00")
						fw_string = fw_data[index_from:index_from+index_to].decode("ASCII", "ignore")
						if "Firmware" in fw_string:
							fw += "<br><br><b>" + __("Detected firmware string:") + "</b><br>" + fw_string
							if "N64 Firmware" in fw_string: raise ValueError(__("JoeyN64 Firmware found"))
							if "Jr4Gen3 Firmware" in fw_string: raise ValueError(__("Jr4Gen3 Firmware found"))
						verified = True
						if len(fw_data) > 0x10000:
							verified = False
							raise ValueError(__("Firmware file is too large."))
					except ValueError as e:
						fw += "<br><br>" + __("Warning: The selected firmware file couldn’t be confirmed to be valid automatically ({error}). Please double check that this is a valid firmware file for your Joey Jr. If it is invalid or an update for a different device, it may render your device unusable.", error=str(e))
					except:
						verified = False

			if verified is False:
				text = __("The firmware update file is corrupted or invalid.")
				self.btnUpdate.setEnabled(True)
				self.btnClose.setEnabled(True)
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
				answer = msgbox.exec()
				return False

			text = __("The following firmware will now be written to your Joey Jr device:") + "\n- " + fw
			text += "\n\n" + __("Do you want to continue?")
			msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Question, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
			msgbox.setDefaultButton(QtWidgets.QMessageBox.Yes)
			answer = msgbox.exec()
			if answer == QtWidgets.QMessageBox.No: return
			self.grpAvailableFwUpdates.setEnabled(False)
			self.btnUpdate.setEnabled(False)
			self.btnClose.setEnabled(False)

			self.APP.DisconnectDevice()

			ret = 0
			while True:
				if ret == 4:
					ret = self.FWUPD.WriteFirmwareMSC(path, fw_data, self.SetStatus)
				else:
					ret = self.FWUPD.WriteFirmware(fw_data, self.SetStatus)

				if ret == 1:
					text = __("The firmware update is complete!")
					if self.optFW_MSC.isChecked():
						text += "\n\n" + __("Please note that you need to manually copy over the {romlist} file. You can find the latest version of it on the Joey Jr website.", romlist="ROMLIST.RAW")
					self.grpAvailableFwUpdates.setEnabled(True)
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Information, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					self.DEVICE = None
					self.reject()
					return True
				elif ret == 2:
					text = __("The firmware update has failed. Please try again.")
					self.grpAvailableFwUpdates.setEnabled(True)
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return False
				elif ret == 3:
					text = __("The firmware update file is corrupted. Please re-install the application.")
					self.grpAvailableFwUpdates.setEnabled(True)
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return False
				elif ret == 4:
					if platform.system() == 'Darwin':
						self.SetStatus(__("No device found."), enableUI=True)
						text = __("If your Joey Jr device is currently running the Drag’n’Drop firmware, please update the firmware on Windows or Linux, or use the standalone firmware updater.")
						msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
						answer = msgbox.exec()
						return False
					answer = QtWidgets.QMessageBox.information(self, AppInfo.NAME, __("If your Joey Jr device is currently running the Drag’n’Drop firmware, please select the <b>{mode_txt}</b> (or <b>{mode_txt_2}</b>) file that is stored on the device.", mode_txt="MODE.TXT", mode_txt_2="MODE!.TXT"), QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel, QtWidgets.QMessageBox.Ok)
					if answer == QtWidgets.QMessageBox.Cancel:
						self.SetStatus(__("No device found."), enableUI=True)
						return False
					path = self.APP.SETTINGS.value("LastDirFirmwareUpdate")
					path = QtWidgets.QFileDialog.getOpenFileName(self, __("Choose the {mode_txt} file of your Joey Jr removable drive", mode_txt="MODE.TXT"), path, "MODE.TXT (MODE*.TXT)")[0]
					self.APP.QT_APP.processEvents()
					if os.path.basename(path) not in ("MODE.TXT", "MODE!.TXT"):
						self.SetStatus(__("No device found."), enableUI=True)
						return False
					self.APP.SETTINGS.setValue("LastDirFirmwareUpdate", os.path.dirname(path))

		def SetStatus(self, text, enableUI=False, setProgress=None):
			self.lblStatus.setText(__("Status: {text}", text=text))
			if setProgress is not None:
				self.prgStatus.setValue(setProgress * 10)
			if enableUI:
				self.grpAvailableFwUpdates.setEnabled(True)
				self.btnUpdate.setEnabled(True)
				self.btnClose.setEnabled(True)
			self.APP.QT_APP.processEvents()
except ImportError:
	pass



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

import os, platform

class DumpReport:
	@classmethod
	def generate(cls, di, device) -> str:
		from .Mapper import DMG_Mapper, ConvertMapperToMapperType
		from .CartridgeTypes import RomSizes, DmgSaveTypes, AgbSaveTypes
		from .Formatter import Formatter
		from .app import AppInfo
		from . import i18n

		def _fields_to_lines(fields, col=19):
			return [f"* {label + ':':<{col}}{value}" for label, value in fields]

		# Resolve header into a shallow copy so we never mutate the caller's dict
		header = dict(di["header"].get("unchanged", di["header"]))
		if "db" in di["header"]:
			header["db"] = di["header"]["db"]

		mode = di["system"]
		if mode not in ("DMG", "AGB"):
			raise NotImplementedError

		system_name = "Game Boy" if mode == "DMG" else "Game Boy Advance"

		rom_size_int = di["rom_size"]
		if rom_size_int in RomSizes():
			rom_size_str = RomSizes(rom_size_int).GetString(localized=False)
		else:
			rom_size_str = f"{rom_size_int:,} bytes"

		keys = list(device.SUPPORTED_CARTS[mode].keys())
		cart_type_str = keys[di["cart_type"]] if 0 <= di["cart_type"] < len(keys) else f"#{di['cart_type']}"
		file_name = os.path.split(di["file_name"])[1] if di["file_name"] else ""
		file_size_bytes = di["file_size"]
		file_size_str = f"{Formatter.file_size(file_size_bytes, space=' ', localized=False)} ({file_size_bytes:d} bytes)"
		logo_str = "OK" if header["logo_correct"] else "Invalid"

		lines = ["= FlashGBX Dump Report ="]

		lines += ["", "== File Information =="]
		lines += _fields_to_lines([
			("File Name",  file_name),
			("File Size",  file_size_str),
			("CRC32",      f"{di['hash_crc32']:08x}"),
			("MD5",        di["hash_md5"]),
			("SHA-1",      di["hash_sha1"]),
			("SHA-256",    di["hash_sha256"]),
		])

		lines += ["", "== General Information =="]
		general_fields = [
			("Hardware",         f"{device.GetFullName()} – Firmware {device.GetFirmwareVersion()}"),
			("Software",         f"{AppInfo.NAME} {AppInfo.VERSION}"),
			("OS Platform",      f"{AppInfo.os_string()}, {platform.machine()}, {i18n.OS_LANGUAGE}"),
		]
		if device.GetName() == "GBxCart RW": general_fields.append(("Baud Rate", f"{device.GetBaudRate():d}"))
		general_fields += [
			("Dump Time",        di["timestamp"]),
			("Time Elapsed",     "%TIME_ELAPSED% (%TRANSFER_RATE%)"),
			("Transfer Buffer",  f"{di['transfer_size']:d} bytes"),
			("Retries",          f"{device.GetReadErrors():d}"),
		]
		lines += _fields_to_lines(general_fields)

		lines += ["", "== Dumping Settings =="]
		dumping_fields = [
			("Mode",     system_name),
			("ROM Size", rom_size_str),
		]
		if mode == "DMG":
			mapper_int = di["mapper_type"]
			if mapper_int in DMG_Mapper().GetAllMapperIds():
				mapper_str = ConvertMapperToMapperType(mapper_int)[0]
			else:
				mapper_str = f"0x{mapper_int:02X}"
			dumping_fields += [
				("Mapper Type",       mapper_str),
				("Cartridge Profile", cart_type_str),
				("Read Method",       di["dmg_read_method"]),
			]
		else:  # AGB
			dumping_fields += [
				("Cartridge Profile", cart_type_str),
				("Read Method",       di["agb_read_method"]),
			]
		lines += _fields_to_lines(dumping_fields)

		lines += ["", "== Parsed Data =="]

		if mode == "DMG":
			cgb_raw = header["cgb"]
			if cgb_raw == 0xC0:
				target_platform = "Game Boy Color exclusive"
			elif cgb_raw == 0x80:
				target_platform = "Game Boy Color"
			elif header["old_lic"] == 0x33 and header["sgb"] == 0x03:
				target_platform = "Super Game Boy"
			else:
				target_platform = "Original Game Boy"

			sgb_str = "Supported" if (header["old_lic"] == 0x33 and header["sgb"] == 0x03) else "No support"
			cgb_str = DMG_Mapper().CGB_MAP.get(cgb_raw, f"Unknown (0x{cgb_raw:02X})")

			hdr_chk = header["header_checksum"]
			hdr_chk_calc = header.get("header_checksum_calc", hdr_chk)
			header_checksum_str = f"OK (0x{hdr_chk:02X})" if header["header_checksum_correct"] \
				else f"Invalid (0x{hdr_chk_calc:02X}≠0x{hdr_chk:02X})"

			header["rom_checksum_calc"] = device.INFO.get("rom_checksum_calc", header.get("rom_checksum_calc"))
			rom_chk_ok = header["rom_checksum_calc"] == header["rom_checksum"]
			rom_checksum_str = f"OK (0x{header['rom_checksum']:04X})" if rom_chk_ok \
				else f"Invalid (0x{header['rom_checksum_calc']:04X}≠0x{header['rom_checksum']:04X})"

			hdr_rom_size_raw = header["rom_size_raw"]
			if hdr_rom_size_raw < RomSizes().GetNumberOfTypes():
				hdr_rom_size_str = RomSizes().GetString(index=hdr_rom_size_raw, localized=False)
			else:
				hdr_rom_size_str = f"Unknown (0x{hdr_rom_size_raw:02X})"

			hdr_save_raw = header["ram_size_raw"]
			if hdr_save_raw == 0x00:
				hdr_save_str = f"No SRAM (0x{hdr_save_raw:02X})"
			elif hdr_save_raw in DmgSaveTypes():
				hdr_save_str = f"{DmgSaveTypes(mbc=hdr_save_raw).GetString(localized=False)} (0x{hdr_save_raw:02X})"
			else:
				hdr_save_str = f"Unknown (0x{hdr_save_raw:02X})"

			mapper_raw = header["mapper_raw"]
			if mapper_raw in DMG_Mapper().GetAllMapperIds():
				hdr_mapper_str = f"{DMG_Mapper().GetMapperName(mapper_raw)} (0x{mapper_raw:02X})"
			else:
				hdr_mapper_str = f"Unknown (0x{mapper_raw:02X})"

			parsed_fields = [("Game Title", (header.get("game_title") or "").replace("\0", "␀"))]
			if cgb_raw in (0xC0, 0x80) and header.get("game_code"):
				parsed_fields.append(("Game Code", header["game_code"].replace("\0", "␀")))
			parsed_fields += [
				("Revision",        str(header["version"])),
				("Super Game Boy",  sgb_str),
				("Game Boy Color",  cgb_str),
				("Nintendo Logo",   logo_str),
				("Header Checksum", header_checksum_str),
				("ROM Checksum",    rom_checksum_str),
				("ROM Size",        hdr_rom_size_str),
				("SRAM Size",       hdr_save_str),
				("Mapper Type",     hdr_mapper_str),
				("Target Platform", target_platform),
			]
			lines += _fields_to_lines(parsed_fields)

			if "gbmem" in di and di["gbmem"] is not None:
				raw_data = "\n                     ".join(
					''.join(f"{x:02X}" for x in di["gbmem"][i*0x20:i*0x20+0x20])
					for i in range(4)
				)
				if "gbmem_parsed" in di and di["gbmem_parsed"] is not None and len(di["gbmem_parsed"]) > 0:
					if isinstance(di["gbmem_parsed"], list):
						p0 = di["gbmem_parsed"][0]
						lines += ["", "== GB-Memory Data (Multi Menu) =="]
						lines += _fields_to_lines([
							("Write Timestamp", p0["timestamp"]),
							("Write Kiosk ID",  p0["kiosk_id"]),
							("Number of Games", f"{p0['num_games']:d}"),
							("Write Counter",   f"{p0['write_count']:d}"),
							("Cartridge ID",    p0["cart_id"]),
							("Raw Map Data",    raw_data),
						])
						for i in range(1, len(di["gbmem_parsed"])):
							entry = di["gbmem_parsed"][i]
							if entry["menu_index"] == 0xFF or not entry["header"]["logo_correct"]:
								continue
							section = "Menu ROM" if i == 1 else f"Game {i - 1}"
							entry_rom_bytes = entry["rom_size"]
							entry_size_str = f"{Formatter.file_size(entry_rom_bytes, space=' ', localized=False)} ({entry_rom_bytes:d} bytes)"
							entry_fields = [
								("Game Code",        entry["game_code"]),
								("Game Title",       entry["title"]),
								("Write Timestamp",  entry["timestamp"]),
								("Write Kiosk ID",   entry["kiosk_id"]),
								("Location",         f"0x{entry['rom_offset']:06X}–0x{entry['rom_offset'] + entry['rom_size'] - 1:06X}"),
								("ROM Size",         entry_size_str),
							]
							if "crc32" in entry: entry_fields.append(("CRC32",    f"{entry['crc32']:08x}"))
							if "md5"   in entry: entry_fields.append(("MD5",      entry["md5"]))
							if "sha1"  in entry: entry_fields.append(("SHA-1",    entry["sha1"]))
							if "sha256" in entry: entry_fields.append(("SHA-256", entry["sha256"]))
							lines += ["", f"=== {section} ==="]
							lines += _fields_to_lines(entry_fields)
							if "db_entry" in entry and "crc32" in entry and entry["db_entry"]["rc"] == entry["crc32"]:
								lines += _fields_to_lines([("Database Match", f"{entry['db_entry']['gn']} {entry['db_entry']['ne']}")])
					elif isinstance(di["gbmem_parsed"]["game_code"], str):
						p = di["gbmem_parsed"]
						lines += ["", "== GB-Memory Data (Single Game) =="]
						lines += _fields_to_lines([
							("Game Code",       p["game_code"]),
							("Game Title",      p["title"]),
							("Write Timestamp", p["timestamp"]),
							("Write Kiosk ID",  p["kiosk_id"]),
							("Write Counter",   f"{p['write_count']:d}"),
							("Cartridge ID",    p["cart_id"]),
							("Raw Map Data",    raw_data),
						])
				else:
					lines += _fields_to_lines([("GB-Memory Data", raw_data)])

			if header["db"] is not None and header["db"]["rc"] == di["hash_crc32"]:
				db = header["db"]
				db_fields = []
				if "gn" in db and "ne" in db: db_fields.append(("Game Name", f"{db['gn']} {db['ne']}"))
				elif "gn" in db: db_fields.append(("Game Name", db["gn"]))
				if "rg" in db: db_fields.append(("Region", db["rg"]))
				if "lg" in db: db_fields.append(("Language(s)", db["lg"]))
				if "rv" in db: db_fields.append(("Revision", db["rv"]))
				if "gc" in db: db_fields.append(("Game Code", db["gc"]))
				if "rc" in db: db_fields.append(("ROM CRC32", f"{db['rc']:08x}"))
				if "rs" in db: db_fields.append(("ROM Size", Formatter.file_size(db["rs"], space=" ", as_int=True, localized=False)))
				lines += ["", "== Database Match =="]
				lines += _fields_to_lines(db_fields)

		elif mode == "AGB":
			hdr_chk = header["header_checksum"]
			hdr_chk_calc = header.get("header_checksum_calc", hdr_chk)
			header_checksum_str = f"OK (0x{hdr_chk:02X})" if header["header_checksum_correct"] \
				else f"Invalid (0x{hdr_chk_calc:02X}≠0x{hdr_chk:02X})"

			savelib_str = AgbSaveTypes().GetStringFromSaveLib(di["agb_savelib"], localized=False)
			game_title_raw = (header.get("game_title_raw") or "").replace("\0", "␀")
			game_code_raw  = (header.get("game_code_raw")  or "").replace("\0", "␀")

			parsed_fields = [
				("Game Title",      game_title_raw),
				("Game Code",       game_code_raw),
				("Revision",        str(header["version"])),
				("Nintendo Logo",   logo_str),
				("Header Checksum", header_checksum_str),
				("Save Type",       savelib_str),
			]
			if "agb_save_flash_id" in di and di["agb_save_flash_id"] is not None:
				chip_id, chip_name = di["agb_save_flash_id"]
				parsed_fields.append(("Save Flash Chip", f"{chip_name} (0x{chip_id:04X})"))
			if "eeprom_data" in di:
				eeprom_hex = ''.join(f"{x:02X}" for x in di["eeprom_data"])
				parsed_fields.append(("EEPROM area", f"{eeprom_hex}"))
			lines += _fields_to_lines(parsed_fields)

			if cart_type_str == "Vast Fame":
				lines += ["", "== Vast Fame Protection Information =="]
				lines += _fields_to_lines([
					("Address Reordering", str(di.get("vf_addr_reorder", "N/A"))),
					("Value Reordering",   str(di.get("vf_value_reorder", "N/A"))),
				], col=21)

			if header["db"] is not None and header["db"]["rc"] == di["hash_crc32"]:
				db = header["db"]
				db_fields = []
				if "gn" in db and "ne" in db: db_fields.append(("Game Name", f"{db['gn']} {db['ne']}"))
				elif "gn" in db: db_fields.append(("Game Name", db["gn"]))
				if "rg" in db: db_fields.append(("Region", db["rg"]))
				if "lg" in db: db_fields.append(("Language(s)", db["lg"]))
				if "rv" in db: db_fields.append(("Revision", db["rv"]))
				if "gc" in db: db_fields.append(("Game Code", db["gc"]))
				if "rc" in db: db_fields.append(("ROM CRC32", f"{db['rc']:08x}"))
				if "rs" in db: db_fields.append(("ROM Size", Formatter.file_size(db["rs"], space=" ", as_int=True, localized=False)))
				if "st" in db: db_fields.append(("Save Type", AgbSaveTypes(db["st"]).GetString(localized=False)))
				lines += ["", "== Database Match =="]
				lines += _fields_to_lines(db_fields)

		newline = "\r\n" if platform.system() == "Windows" else "\n"
		return newline.join(lines)



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

# pylint: disable=wildcard-import, unused-wildcard-import
import datetime, hashlib, math, os, random, re, struct, time, zipfile
from serial import SerialException
from .app import AppInfo
from .LK_Device import *
from .i18n import __, c__, format_decimal
from .IniSettings import IniSettings

class GbxDevice(LK_Device):
	DEVICE_NAME = "GBxCart RW"
	DEVICE_MIN_FW = 1
	DEVICE_MAX_FW = 1
	DEVICE_LATEST_FW_TS = { 4:1709317610, 5:1780508702, 6:1780508702, 2:0, 90:0, 100:0 }
	PCB_VERSIONS = { 5:'v1.4', 6:'v1.4a/b/c', 2:'v1.1/v1.2', 4:'v1.3', 90:'XMAS v1.0', 100:'Mini v1.0' }
	DEVICE_LABEL_LONG = "GBxCart RW v1.4 or v1.4a/b/c"
	DEVICE_LABEL_SHORT = "GBxCart RW v1.4"
	FWUPDATE_ACTION = "fwupdate-gbxcartrw"
	CLI_UPDATER_METHOD = "UpdateFirmwareGBxCartRW"
	DEVICE_SUPPORT_MESSAGE = "For help with your GBxCart RW, please visit the insideGadgets Discord:\nhttps://gbxcart.com/discord"

	BAUDRATE = 1000000
	MAX_BUFFER_READ = 0x1000
	MAX_BUFFER_WRITE = 0x400
	DEVICE_CMD = LK_Device.DEVICE_CMD.copy()
	DEVICE_CMD.update({
		"OFW_RESET_AVR":0x2A,
		"OFW_CART_MODE":0x43,
		"OFW_FW_VER":0x56,
		"OFW_PCB_VER":0x68,
		"OFW_USART_1_0M_SPEED":0x3C,
		"OFW_USART_1_5M_SPEED":0x3E,
		"OFW_CART_PWR_ON":0x2F,
		"OFW_CART_PWR_OFF":0x2E,
		"OFW_QUERY_CART_PWR":0x5D,
		"OFW_DONE_LED_ON":0x3D,
		"OFW_ERROR_LED_ON":0x3F,
		"OFW_GB_CART_MODE":0x47,
		"OFW_GB_FLASH_BANK_1_COMMAND_WRITES":0x4E,
		"OFW_LNL_QUERY":0x25,
	})

	def Initialize(self, flashcarts=None, port=None, max_baud=1500000):
		if self.IsConnected(): self.DEVICE.close()
		if platform.system() == "Darwin": max_baud = 1000000

		conn_msg = []
		ports = []
		if port is not None:
			ports = [ port ]
		else:
			comports = serial.tools.list_ports.comports()
			for i in range(0, len(comports)):
				if comports[i].vid == 0x1A86 and comports[i].pid == 0x7523:
					ports.append(comports[i].device)
			if len(ports) == 0: return False

		for i in range(0, len(ports)):
			for baudrate in (1000000, 1500000):
				if max_baud < baudrate: continue
				try:
					if self.TryConnect(ports[i], baudrate):
						self.BAUDRATE = baudrate
						dev = serial.Serial(ports[i], self.BAUDRATE, timeout=0.1)
						self.DEVICE = dev
						break
				except SerialException as e:
					if "Permission" in str(e):
						conn_msg.append([3, __("The device on port {port} couldn’t be accessed. Make sure your user account has permission to use it and it’s not already in use by another application.", port=ports[i])])
					elif "FileNotFoundError" in str(e):
						continue
					else:
						conn_msg.append([3, __("A critical error occured while trying to access the device on port {port}.", port=ports[i]) + "\n\n" + str(e)])
					continue

			if self.FW is None or self.FW == {}: continue
			if max_baud >= 1500000 and self.FW is not None and "pcb_ver" in self.FW and self.FW["pcb_ver"] in (5, 6, 101) and self.BAUDRATE < 1500000:
				self.ChangeBaudRate(baudrate=1500000)
				self.DEVICE.close()
				dev = serial.Serial(ports[i], self.BAUDRATE, timeout=0.1)
				self.DEVICE = dev

			dprint(f"Found a {self.DEVICE_NAME}")
			dprint("Firmware information:", self.FW)
			dprint("Baud rate:", self.BAUDRATE)

			if self.DEVICE is None or not self.IsConnected() or self.FW is None or self.FW["pcb_ver"] not in self.DEVICE_LATEST_FW_TS:
				self.DEVICE = None
				if self.FW is not None:
					conn_msg.append([0, __("Couldn’t communicate with the {device_name} on port {port}. Please disconnect and reconnect the device, then try again.", device_name=self.DEVICE_NAME, port=ports[i])])
				continue
			elif self.FW["fw_ts"] > self.DEVICE_LATEST_FW_TS[self.FW["pcb_ver"]]:
				conn_msg.append([1, __("Note: The {device_name} on port {port} is running a firmware version that is newer than what this version of FlashGBX was developed to work with, so errors may occur.", device_name=self.DEVICE_NAME, port=ports[i])])
			elif self.FW["pcb_ver"] in (5, 6, 101) and self.BAUDRATE > 1000000:
				self.MAX_BUFFER_READ = 0x1000
				self.MAX_BUFFER_WRITE = 0x400
			else:
				self.MAX_BUFFER_READ = 0x1000
				self.MAX_BUFFER_WRITE = 0x100

			self.PORT = ports[i]
			self.DEVICE.timeout = self.DEVICE_TIMEOUT

			# Load Flash Cartridge Handlers
			if flashcarts is not None:
				self.UpdateFlashCarts(flashcarts)

			# Stop after first found device
			break

		return conn_msg

	def LoadFirmwareVersion(self):
		dprint("Querying firmware version")
		try:
			self.DEVICE.timeout = 0.075
			self.DEVICE.reset_input_buffer()
			self.DEVICE.reset_output_buffer()
			self._write(self.DEVICE_CMD["OFW_PCB_VER"])
			temp = self.DEVICE.read(1)
			self.DEVICE.timeout = self.DEVICE_TIMEOUT
			if len(temp) == 0:
				dprint("No response")
				self.FW = None
				return False
			pcb = temp[0]
			if pcb == b'': return False
			self._write(self.DEVICE_CMD["OFW_FW_VER"])
			ofw = self._read(1)
			if (pcb == 2 and ofw == 2):
				dprint(f"Not a {self.DEVICE_NAME}")
				self.FW = None
				return False
			if (pcb >= 5 and ofw == 0):
				dprint(f"Not a {self.DEVICE_NAME}")
				self.FW = None
				return False
			if (pcb < 5 and ofw > 0):
				self.FW = {
					"ofw_ver":ofw,
					"pcb_ver":pcb,
					"pcb_name":"GBxCart RW",
					"cfw_id":"",
					"fw_ver":0,
					"fw_ts":0,
					"fw_dt":"",
				}
				return True

			self._write(self.DEVICE_CMD["QUERY_FW_INFO"])
			size = self._read(1)
			if size != 8: return False
			data = self._read(size)
			info = data[:8]
			keys = ["cfw_id", "fw_ver", "pcb_ver", "fw_ts"]
			values = struct.unpack(">cHBI", bytearray(info))
			self.FW = dict(zip(keys, values))
			self.FW["cfw_id"] = self.FW["cfw_id"].decode('ascii')
			self.FW["fw_dt"] = datetime.datetime.fromtimestamp(self.FW["fw_ts"]).astimezone().replace(microsecond=0).isoformat()
			self.FW["ofw_ver"] = ofw
			self.FW["pcb_name"] = ""
			self.FW["cart_power_ctrl"] = False
			self.FW["bootloader_reset"] = False
			if self.FW["cfw_id"] == "L" and self.FW["fw_ver"] >= 12:
				size = self._read(1)
				name = self._read(size)
				if len(name) > 0:
					try:
						self.FW["pcb_name"] = name.decode("UTF-8").replace("\x00", "").strip()
					except:
						self.FW["pcb_name"] = "Unnamed Device"
					self.DEVICE_NAME = self.FW["pcb_name"]

				# Cartridge Power Control support
				temp = self._read(1)
				self.FW["cart_power_ctrl"] = True if temp & 1 == 1 else False
				self.FW["cart_presence_switch"] = True if (temp >> 1) & 1 == 1 else False
				self.FW["cart_mode_switch"] = True if (temp >> 2) & 1 == 1 else False

				# Reset to bootloader support
				self.FW["bootloader_reset"] = True if self._read(1) == 1 else False

			return True

		except Exception as e:
			dprint("Disconnecting due to an error", e, sep="\n")
			try:
				if self.DEVICE.isOpen():
					self.DEVICE.reset_input_buffer()
					self.DEVICE.reset_output_buffer()
					self.DEVICE.close()
				self.DEVICE = None
			except:
				pass
			return False

	def ChangeBaudRate(self, baudrate):
		if not self.IsConnected(): return
		dprint("Changing baud rate to", baudrate)
		if baudrate == 1500000:
			self._write(self.DEVICE_CMD["OFW_USART_1_5M_SPEED"])
		elif baudrate == 1000000:
			self._write(self.DEVICE_CMD["OFW_USART_1_0M_SPEED"])
		self.BAUDRATE = baudrate
		self.DEVICE.close()

	def CheckActive(self):
		if time.time() < self.LAST_CHECK_ACTIVE + 1: return True
		dprint("Checking if device is active (GBxCart RW specific)")
		if self.DEVICE is None: return False
		if self.FW["pcb_name"] is None:
			if self.LoadFirmwareVersion():
				self.LAST_CHECK_ACTIVE = time.time()
				return True
			return False
		try:
			if self.FW["fw_ver"] == 0: # legacy GBxCart RW firmware
				self.LAST_CHECK_ACTIVE = time.time()
				return True
			elif self.FW["fw_ver"] < 12:
				self._write(bytearray([self.DEVICE_CMD["OFW_FW_VER"]]))
				self._read(1)
				self.LAST_CHECK_ACTIVE = time.time()
				return True
			else:
				return super().CheckActive()
		except Exception as e:
			dprint("Disconnecting...", e)
			try:
				if self.DEVICE.isOpen():
					self.DEVICE.reset_input_buffer()
					self.DEVICE.reset_output_buffer()
					self.DEVICE.close()
				self.DEVICE = None
			except:
				pass
			return False

	def GetFirmwareVersion(self, more=False):
		if self.FW["fw_ver"] == 0: # old GBxCart RW
			return "R{:d}".format(self.FW["ofw_ver"])

		if self.FW["pcb_ver"] in (5, 6, 101):
			s = "R{:d}+{:s}{:d}".format(self.FW["ofw_ver"], self.FW["cfw_id"], self.FW["fw_ver"])
		else:
			s = "{:s}{:d}".format(self.FW["cfw_id"], self.FW["fw_ver"])
		if more:
			s += " ({:s})".format(self.FW["fw_dt"])
		return s

	def GetFullNameExtended(self, more=False):
		if self.FW["fw_ver"] == 0:  # old GBxCart RW
			return __("{device_name} – Firmware {fw_version} ({port})", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), port=self.GetPort())

		if more:
			return __("{device_name} – Firmware {fw_version} ({timestamp}) on {port} at {baudrate}M baud", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), timestamp=self.FW["fw_dt"], port=self.GetPort(), baudrate=format_decimal(self.BAUDRATE/1000/1000, precision=1))
		else:
			return __("{device_name} – Firmware {fw_version} ({port})", device_name=self.GetFullName(), fw_version=self.GetFirmwareVersion(), port=self.GetPort())

	def CanSetVoltageBySwitch(self):
		return False

	def CanSetVoltageByCode(self):
		return True

	def CanSetVoltageByAutoswitch(self):
		return False

	def CanPowerCycleCart(self):
		if self.FW is None or self.DEVICE is None: return False
		if not self.DEVICE.is_open: return False
		if self.FW["fw_ver"] >= 12:
			return self.FW["cart_power_ctrl"]
		else:
			return self.FW["pcb_ver"] in (5, 6)

	def GetSupprtedModes(self):
		if self.FW["pcb_ver"] == 101:
			return ["DMG"]
		else:
			return ["DMG", "AGB"]

	def IsSupported3dMemory(self):
		return True

	def IsClkConnected(self):
		return self.FW["pcb_ver"] in (5, 6, 101)

	def SupportsFirmwareUpdates(self):
		if not isinstance(getattr(self, "FW", None), dict):
			return True
		if "pcb_ver" not in self.FW:
			return True
		if "ofw_ver" not in self.FW:
			return self.FW["pcb_ver"] in (2, 4, 5, 6, 90, 100, 101)

		if self.FW["ofw_ver"] == 30:
			if self.DEVICE is not None:
				self._write(self.DEVICE_CMD["OFW_LNL_QUERY"])
				old_timeout = self.DEVICE.timeout
				self.DEVICE.timeout = 0.15
				is_lnl = self._read(1) == 0x31
				self.DEVICE.timeout = old_timeout
				dprint("LinkNLoad detected:", is_lnl)
				if is_lnl: return False
		return self.FW["pcb_ver"] in (2, 4, 5, 6, 90, 100, 101)

	def FirmwareUpdateAvailable(self):
		if self.FW["fw_ver"] == 0 and self.FW["pcb_ver"] in (2, 4, 90, 100, 101):
			if self.FW["pcb_ver"] == 4:
				self.FW_UPDATE_REQ = True
			else:
				self.FW_UPDATE_REQ = 2
			return True
		if self.FW["pcb_ver"] not in (4, 5, 6): return False
		if self.FW["pcb_ver"] in (5, 6) and self.FW["fw_ts"] != self.DEVICE_LATEST_FW_TS[self.FW["pcb_ver"]]:
			return True
		if self.FW["pcb_ver"] == 4 and self.FW["fw_ts"] != self.DEVICE_LATEST_FW_TS[self.FW["pcb_ver"]]:
			self.FW_UPDATE_REQ = True
			return True

	def GetFirmwareUpdaterClass(self):
		if self is None or self.FW["pcb_ver"] in (5, 6): # v1.4 / v1.4a/b/c
			try:
				return (FirmwareUpdater, FirmwareUpdaterWindow)
			except:
				return None
		elif self.FW["pcb_ver"] in (2, 4, 90, 100, 101): # v1.3
			try:
				return (None, FirmwareUpdaterWindowV13)
			except:
				return None
		else:
			return None

	def ResetLEDs(self):
		if self.DEVICE in (None, False): return
		self._write(self.DEVICE_CMD["OFW_CART_MODE"]) # Reset LEDs
		self._read(1)

	def SupportsBootloaderReset(self):
		if self.FW["fw_ver"] >= 12:
			return self.FW["bootloader_reset"]
		else:
			return False

	def BootloaderReset(self):
		return False

	def SupportsAudioAsWe(self):
		return True

	def Close(self, cartPowerOff=False):
		try:
			self.ResetLEDs()
		except SerialException:
			pass
		return super().Close(cartPowerOff)

	def SetTimeout(self, seconds=1):
		if seconds < 1: seconds = 1
		self.DEVICE_TIMEOUT = seconds
		self.DEVICE.timeout = self.DEVICE_TIMEOUT


class FirmwareUpdater():
	PORT = ""

	def __init__(self, app_path=".", port=None):
		self.APP_PATH = app_path
		self.PORT = port

	def WriteFirmware(self, zipfn, fncSetStatus):
		try:
			with zipfile.ZipFile(zipfn) as archive:
				with archive.open("fw.ini") as f: buffer1 = bytearray(f.read())
				with archive.open("fw.bin") as f: buffer2 = bytearray(f.read())
		except (zipfile.BadZipFile, KeyError):
			fncSetStatus(__("The firmware update file is corrupted."))
			return 3
		if len(buffer2) < 0x20:
			fncSetStatus(__("The firmware update file is corrupted."))
			return 3
		while len(buffer1) < len(buffer2): buffer1 = buffer1 + buffer1
		random.seed(struct.unpack("<I", buffer2[-0x18:-0x14])[0])
		chk = buffer2[-0x14:]
		buffer = bytearray()
		for i in range(0, len(buffer2[0:-0x18])):
			r = int(random.random()*256) % 256
			buffer.append(buffer2[0:-0x18][len(buffer2[0:-0x18]) - i - 1] ^ r ^ buffer1[len(buffer1) - i - 1])
		if (chk != hashlib.sha1(buffer).digest()):
			fncSetStatus(__("The firmware update file is corrupted."))
			return 3

		if self.PORT is None:
			ports = []
			comports = serial.tools.list_ports.comports()
			for i in range(0, len(comports)):
				if comports[i].vid == 0x1A86 and comports[i].pid == 0x7523:
					ports.append(comports[i].device)
			if len(ports) == 0:
				fncSetStatus(__("No device found."))
				return 2
			port = ports[0]
		else:
			port = self.PORT
		data = buffer
		buffer = bytearray()

		fncSetStatus(text=__("Connecting..."))
		try:
			dev = serial.Serial(port=port, baudrate=57600, timeout=1)
		except:
			fncSetStatus(text=__("Device not accessible."), enableUI=True)
			return 2
		dev.reset_input_buffer()

		# Write firmware
		fncSetStatus(__("Updating firmware..."), setProgress=0)

		size = len(data)
		counter = 0
		while counter < size:
			byte = data[counter:counter+1]
			dev.write(byte)
			tmp_byte = dev.read(1)
			if (tmp_byte != byte):
				try:
					tmp_byte = int.from_bytes(tmp_byte, byteorder="little")
				except:
					tmp_byte = 0
				byte = int.from_bytes(byte, byteorder="little")
				if counter == 0:
					fncSetStatus(text=__("Update failed!"), enableUI=True)
				else:
					fncSetStatus(text=__("Update failed at offset {offset}!", offset="0x{:04X}".format(counter)), enableUI=True)
				return 2

			counter += 1
			percent = float(counter)/size*100
			fncSetStatus(text=__("Updating firmware... Do not unplug the device!"), setProgress=percent)

		dev.close()
		time.sleep(0.8)
		fncSetStatus(__("Done!"))
		time.sleep(0.2)
		return 1


try:
	from .pyside import QtCore, QtWidgets, QtGui, QDesktopWidget

	class FirmwareUpdaterWindow(QtWidgets.QDialog):
		APP = None
		DEVICE = None
		FWUPD = None
		DEV_NAME = "GBxCart RW"
		FW_VER = ""
		PCB_VER = ""

		def __init__(self, app, app_path, file=None, icon=None, device=None):
			QtWidgets.QDialog.__init__(self, app)
			if icon is not None: self.setWindowIcon(QtGui.QIcon(icon))
			self.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")
			self.setWindowTitle(AppInfo.NAME + " – " + __("Firmware Updater for {device_name}", device_name="GBxCart RW"))
			self.setWindowFlags((self.windowFlags() | QtCore.Qt.MSWindowsFixedSizeDialogHint) & ~QtCore.Qt.WindowContextHelpButtonHint)

			self.APP = app
			if device is not None:
				self.FWUPD = FirmwareUpdater(app_path, device.GetPort())
				self.DEV_NAME = device.GetName()
				self.FW_VER = device.GetFirmwareVersion(more=True)
				self.PCB_VER = device.GetPCBVersion()
				self.DEVICE = device
			else:
				self.APP.QT_APP.processEvents()
				self.FWUPD = FirmwareUpdater(app_path, None)

			self.layout = QtWidgets.QGridLayout()
			self.layout.setContentsMargins(-1, 8, -1, 8)
			self.layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
			self.layout_device = QtWidgets.QVBoxLayout()

			# ↓↓↓ Current Device Information
			self.grpDeviceInfo = QtWidgets.QGroupBox(__("Current Firmware"))
			self.grpDeviceInfo.setMinimumWidth(420)
			self.grpDeviceInfoLayout = QtWidgets.QVBoxLayout()
			self.grpDeviceInfoLayout.setContentsMargins(-1, 3, -1, -1)
			rowDeviceInfo1 = QtWidgets.QHBoxLayout()
			self.lblDeviceName = QtWidgets.QLabel(__("Device:"))
			self.lblDeviceName.setMinimumWidth(120)
			self.lblDeviceNameResult = QtWidgets.QLabel("GBxCart RW")
			rowDeviceInfo1.addWidget(self.lblDeviceName)
			rowDeviceInfo1.addWidget(self.lblDeviceNameResult)
			rowDeviceInfo1.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo1)
			rowDeviceInfo3 = QtWidgets.QHBoxLayout()
			self.lblDeviceFWVer = QtWidgets.QLabel(__("Firmware version:"))
			self.lblDeviceFWVer.setMinimumWidth(120)
			self.lblDeviceFWVerResult = QtWidgets.QLabel("")
			rowDeviceInfo3.addWidget(self.lblDeviceFWVer)
			rowDeviceInfo3.addWidget(self.lblDeviceFWVerResult)
			rowDeviceInfo3.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo3)
			rowDeviceInfo2 = QtWidgets.QHBoxLayout()
			self.lblDevicePCBVer = QtWidgets.QLabel(__("PCB version:"))
			self.lblDevicePCBVer.setMinimumWidth(120)
			self.optDevicePCBVer14 = QtWidgets.QRadioButton("v1.4")
			self.connect(self.optDevicePCBVer14, QtCore.SIGNAL("clicked()"), self.SetPCBVersion)
			self.optDevicePCBVer14a = QtWidgets.QRadioButton("v1.4a/b/c")
			self.connect(self.optDevicePCBVer14a, QtCore.SIGNAL("clicked()"), self.SetPCBVersion)
			rowDeviceInfo2.addWidget(self.lblDevicePCBVer)
			rowDeviceInfo2.addWidget(self.optDevicePCBVer14)
			rowDeviceInfo2.addWidget(self.optDevicePCBVer14a)
			rowDeviceInfo2.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo2)
			self.grpDeviceInfo.setLayout(self.grpDeviceInfoLayout)
			self.layout_device.addWidget(self.grpDeviceInfo)
			# ↑↑↑ Current Device Information

			# ↓↓↓ Available Firmware Updates
			self.grpAvailableFwUpdates = QtWidgets.QGroupBox(__("Available Firmware"))
			self.grpAvailableFwUpdates.setMinimumWidth(400)
			self.grpAvailableFwUpdatesLayout = QtWidgets.QVBoxLayout()
			self.grpAvailableFwUpdatesLayout.setContentsMargins(-1, 3, -1, -1)

			rowDeviceInfo4 = QtWidgets.QHBoxLayout()
			self.lblDeviceFWVer2 = QtWidgets.QLabel(__("Firmware version:"))
			self.lblDeviceFWVer2.setMinimumWidth(120)
			self.lblDeviceFWVer2Result = QtWidgets.QLabel("(" + __("Please choose the PCB version") + ")")
			rowDeviceInfo4.addWidget(self.lblDeviceFWVer2)
			rowDeviceInfo4.addWidget(self.lblDeviceFWVer2Result)
			rowDeviceInfo4.addStretch(1)
			self.grpAvailableFwUpdatesLayout.addLayout(rowDeviceInfo4)

			self.rowUpdate = QtWidgets.QHBoxLayout()
			self.btnUpdate = QtWidgets.QPushButton(__("Install Firmware Update"))
			self.btnUpdate.setMinimumWidth(200)
			self.btnUpdate.setContentsMargins(20, 20, 20, 20)
			self.connect(self.btnUpdate, QtCore.SIGNAL("clicked()"), lambda: [ self.UpdateFirmware() ])
			self.rowUpdate.addStretch()
			self.rowUpdate.addWidget(self.btnUpdate)
			self.rowUpdate.addStretch()

			self.grpAvailableFwUpdatesLayout.addSpacing(3)
			self.grpAvailableFwUpdatesLayout.addItem(self.rowUpdate)
			self.grpAvailableFwUpdates.setLayout(self.grpAvailableFwUpdatesLayout)
			self.layout_device.addWidget(self.grpAvailableFwUpdates)
			# ↑↑↑ Available Firmware Updates

			self.grpStatus = QtWidgets.QGroupBox("")
			self.grpStatusLayout = QtWidgets.QGridLayout()
			self.prgStatus = QtWidgets.QProgressBar()
			self.prgStatus.setMinimum(0)
			self.prgStatus.setMaximum(1000)
			self.prgStatus.setValue(0)
			self.lblStatus = QtWidgets.QLabel(__("Status: Ready."))

			self.grpStatusLayout.addWidget(self.prgStatus, 1, 0)
			self.grpStatusLayout.addWidget(self.lblStatus, 2, 0)

			self.grpStatus.setLayout(self.grpStatusLayout)
			self.layout_device.addWidget(self.grpStatus)

			self.grpFooterLayout = QtWidgets.QHBoxLayout()
			self.btnClose = QtWidgets.QPushButton(c__("Button (& = Keyboard Shortcut)", "&Close"))
			self.connect(self.btnClose, QtCore.SIGNAL("clicked()"), lambda: [ self.reject() ])
			self.grpFooterLayout.addStretch()
			self.grpFooterLayout.addWidget(self.btnClose)
			self.layout_device.addItem(self.grpFooterLayout)

			self.layout.addLayout(self.layout_device, 0, 0)
			self.setLayout(self.layout)

			self.lblDeviceNameResult.setText(self.DEV_NAME)
			self.lblDeviceFWVerResult.setText(self.FW_VER)
			if self.PCB_VER == "v1.4":
				self.optDevicePCBVer14.setChecked(True)
				self.optDevicePCBVer14a.setEnabled(False)
			elif self.PCB_VER == "v1.4a/b/c":
				self.optDevicePCBVer14a.setChecked(True)
				self.optDevicePCBVer14.setEnabled(False)
			self.SetPCBVersion()

		def SetPCBVersion(self):
			if self.optDevicePCBVer14.isChecked():
				file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_GBxCart_RW_v1_4.zip")
			elif self.optDevicePCBVer14a.isChecked():
				file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_GBxCart_RW_v1_4a.zip")
			else:
				return

			with zipfile.ZipFile(file_name) as zip:
				with zip.open("fw.ini") as f: ini_file = f.read()
				ini_file = ini_file.decode(encoding="utf-8")
				self.INI = IniSettings(ini=ini_file, main_section="Firmware")
				self.OFW_VER = self.INI.GetValue("fw_ver")
				self.OFW_BUILDTS = self.INI.GetValue("fw_buildts")
				self.OFW_TEXT = self.INI.GetValue("fw_text")

			self.lblDeviceFWVer2Result.setText("{:s} ({:s})".format(self.OFW_VER, datetime.datetime.fromtimestamp(int(self.OFW_BUILDTS)).astimezone().replace(microsecond=0).isoformat()))

		def run(self):
			try:
				self.layout.update()
				self.layout.activate()
				screenGeometry = QDesktopWidget().screenGeometry(self)
				x = (screenGeometry.width() - self.width()) / 2
				y = (screenGeometry.height() - self.height()) / 2
				self.move(x, y)
				self.show()
			except:
				return

		def hideEvent(self, event):
			if self.DEVICE is None:
				self.APP.ConnectDevice()
			self.APP.activateWindow()

		def reject(self):
			if self.CloseDialog():
				super().reject()

		def CloseDialog(self):
			if self.btnClose.isEnabled() is False:
				text = __("<b>Warning:</b> If you close this window while a firmware update is still running, it might leave the device in an unbootable state.") + " " + __("You can still recover it by running the Firmware Updater again later.") + "<br><br>" + __("Are you sure you want to close this window?")
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Warning, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
				msgbox.setDefaultButton(QtWidgets.QMessageBox.No)
				answer = msgbox.exec()
				if answer == QtWidgets.QMessageBox.No: return False
			return True

		def UpdateFirmware(self):
			if self.optDevicePCBVer14.isChecked():
				device_version = "v1.4"
				file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_GBxCart_RW_v1_4.zip")
				led = "Done"
			elif self.optDevicePCBVer14a.isChecked():
				device_version = "v1.4a/b/c"
				file_name = self.FWUPD.APP_PATH + os.sep + os.path.join("res", "fw_GBxCart_RW_v1_4a.zip")
				led = "Status"
			else:
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=__("Please select the PCB version of your GBxCart RW device."), standardButtons=QtWidgets.QMessageBox.Ok)
				answer = msgbox.exec()
				return False

			self.APP.DisconnectDevice()

			text = __("Please follow these steps to proceed with the firmware update:")
			text += "\n\n" + __(
				"- Disconnect the USB cable of your GBxCart RW {device_version}.\n"
				"- On the circuit board of your GBxCart RW {device_version}, press and hold down the small button while connecting the USB cable again.\n"
				"- Keep the small button held for at least 2 seconds, then let go of it.\n"
				"- If done right, the green LED labeled “{led}” should remain lit.",
				device_version=device_version,
				led=led
			)
			text += "\n" + __("- Click OK to continue.")
			msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Information, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
			msgbox.setDefaultButton(QtWidgets.QMessageBox.Ok)
			answer = msgbox.exec()
			if answer == QtWidgets.QMessageBox.Cancel: return
			self.btnUpdate.setEnabled(False)
			self.btnClose.setEnabled(False)
			self.optDevicePCBVer14.setEnabled(False)
			self.optDevicePCBVer14a.setEnabled(False)

			while True:
				ret = self.FWUPD.WriteFirmware(file_name, self.SetStatus)
				if ret == 1:
					text = __("The firmware update is complete!")
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					self.optDevicePCBVer14.setEnabled(True)
					self.optDevicePCBVer14a.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Information, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					self.DEVICE = None
					self.reject()
					return True
				elif ret == 2:
					text = __("The firmware update has failed. Please try again.")
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					self.optDevicePCBVer14.setEnabled(True)
					self.optDevicePCBVer14a.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return False
				elif ret == 3:
					text = __("The firmware update file is corrupted. Please re-install the application.")
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					self.optDevicePCBVer14.setEnabled(True)
					self.optDevicePCBVer14a.setEnabled(True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return False

		def SetStatus(self, text, enableUI=False, setProgress=None):
			self.lblStatus.setText(__("Status: {text}", text=text))
			if setProgress is not None:
				self.prgStatus.setValue(setProgress * 10)
			if enableUI:
				self.btnUpdate.setEnabled(True)
				self.btnClose.setEnabled(True)
				self.optDevicePCBVer14.setEnabled(True)
				self.optDevicePCBVer14a.setEnabled(True)
			self.APP.QT_APP.processEvents()
except ImportError:
	pass


try:
	from .pyside import QtCore, QtWidgets, QtGui, QDesktopWidget

	class FirmwareUpdaterWindowV13(QtWidgets.QDialog):
		APP = None
		DEVICE = None
		PORT = ""
		FW_FILES = {"v1.1/v1.2":"fw_GBxCart_RW_v1_1_v1_2.zip", "v1.3":"fw_GBxCart_RW_v1_3.zip", "XMAS v1.0":"fw_GBxCart_RW_XMAS_v1_0.zip", "Mini v1.0":"fw_GBxCart_RW_Mini_v1_0.zip"}

		def __init__(self, app, app_path, file=None, icon=None, device=None):
			QtWidgets.QDialog.__init__(self, app)
			if icon is not None: self.setWindowIcon(QtGui.QIcon(icon))
			self.setStyleSheet("QMessageBox { messagebox-text-interaction-flags: 5; }")
			self.APP = app
			self.APP_PATH = app_path
			self.DEVICE = device
			self.PCB_VER = device.GetPCBVersion()
			self.FW_VER = device.GetFirmwareVersion()
			self.PORT = device.GetPort()

			self.setWindowTitle(AppInfo.NAME + " – " + __("Firmware Updater for {device_name}", device_name="GBxCart RW"))
			self.setWindowFlags((self.windowFlags() | QtCore.Qt.MSWindowsFixedSizeDialogHint) & ~QtCore.Qt.WindowContextHelpButtonHint)

			with zipfile.ZipFile(self.APP_PATH + os.sep + os.path.join("res", "{:s}".format(self.FW_FILES[self.PCB_VER]))) as zip:
				with zip.open("fw.ini") as f: ini_file = f.read()
				ini_file = ini_file.decode(encoding="utf-8")
				self.INI = IniSettings(ini=ini_file, main_section="Firmware")
				self.CFW_VER = self.INI.GetValue("cfw_ver")
				self.CFW_TEXT = self.INI.GetValue("cfw_text")
				self.OFW_VER = self.INI.GetValue("ofw_ver")
				self.OFW_TEXT = self.INI.GetValue("ofw_text")

			self.layout = QtWidgets.QGridLayout()
			self.layout.setContentsMargins(-1, 8, -1, 8)
			self.layout.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
			self.layout_device = QtWidgets.QVBoxLayout()

			# ↓↓↓ Current Device Information
			self.grpDeviceInfo = QtWidgets.QGroupBox(__("Current Device Information"))
			self.grpDeviceInfo.setMinimumWidth(420)
			self.grpDeviceInfoLayout = QtWidgets.QVBoxLayout()
			self.grpDeviceInfoLayout.setContentsMargins(-1, 3, -1, -1)
			rowDeviceInfo1 = QtWidgets.QHBoxLayout()
			self.lblDeviceName = QtWidgets.QLabel(__("Device Name:"))
			self.lblDeviceName.setMinimumWidth(120)
			self.lblDeviceNameResult = QtWidgets.QLabel("GBxCart RW")
			rowDeviceInfo1.addWidget(self.lblDeviceName)
			rowDeviceInfo1.addWidget(self.lblDeviceNameResult)
			rowDeviceInfo1.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo1)
			rowDeviceInfo2 = QtWidgets.QHBoxLayout()
			self.lblDevicePCBVer = QtWidgets.QLabel(__("PCB version:"))
			self.lblDevicePCBVer.setMinimumWidth(120)
			self.lblDevicePCBVerResult = QtWidgets.QLabel("1.3")
			rowDeviceInfo2.addWidget(self.lblDevicePCBVer)
			rowDeviceInfo2.addWidget(self.lblDevicePCBVerResult)
			rowDeviceInfo2.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo2)
			rowDeviceInfo3 = QtWidgets.QHBoxLayout()
			self.lblDeviceFWVer = QtWidgets.QLabel(__("Firmware version:"))
			self.lblDeviceFWVer.setMinimumWidth(120)
			self.lblDeviceFWVerResult = QtWidgets.QLabel("R26")
			rowDeviceInfo3.addWidget(self.lblDeviceFWVer)
			rowDeviceInfo3.addWidget(self.lblDeviceFWVerResult)
			rowDeviceInfo3.addStretch(1)
			self.grpDeviceInfoLayout.addLayout(rowDeviceInfo3)
			self.grpDeviceInfo.setLayout(self.grpDeviceInfoLayout)
			self.layout_device.addWidget(self.grpDeviceInfo)
			# ↑↑↑ Current Device Information

			# ↓↓↓ Available Firmware Updates
			self.grpAvailableFwUpdates = QtWidgets.QGroupBox(__("Firmware Update Options"))
			self.grpAvailableFwUpdates.setMinimumWidth(400)
			self.grpAvailableFwUpdatesLayout = QtWidgets.QVBoxLayout()
			self.grpAvailableFwUpdatesLayout.setContentsMargins(-1, 3, -1, -1)

			self.optCFW = QtWidgets.QRadioButton("{:s}".format(self.CFW_VER))
			self.lblCFW_Info = QtWidgets.QLabel("{:s}".format(self.CFW_TEXT))
			self.lblCFW_Info.setWordWrap(True)
			self.lblCFW_Info.mousePressEvent = lambda x: [ self.optCFW.setChecked(True) ]
			self.optOFW = QtWidgets.QRadioButton("{:s}".format(self.OFW_VER))
			self.lblOFW_Info = QtWidgets.QLabel("{:s}".format(self.OFW_TEXT))
			self.lblOFW_Info.setWordWrap(True)
			self.lblOFW_Info.mousePressEvent = lambda x: [ self.optOFW.setChecked(True) ]
			self.optExternal = QtWidgets.QRadioButton(__("External firmware file"))

			self.rowUpdate = QtWidgets.QHBoxLayout()
			self.btnUpdate = QtWidgets.QPushButton(__("Install Firmware Update"))
			self.btnUpdate.setMinimumWidth(200)
			self.btnUpdate.setContentsMargins(20, 20, 20, 20)
			self.connect(self.btnUpdate, QtCore.SIGNAL("clicked()"), lambda: [ self.UpdateFirmware() ])
			self.rowUpdate.addStretch()
			self.rowUpdate.addWidget(self.btnUpdate)
			self.rowUpdate.addStretch()

			if self.PCB_VER == "v1.3":
				self.grpAvailableFwUpdatesLayout.addWidget(self.optCFW)
				self.grpAvailableFwUpdatesLayout.addWidget(self.lblCFW_Info)
				self.optCFW.setChecked(True)
			else:
				self.optOFW.setChecked(True)
			self.grpAvailableFwUpdatesLayout.addWidget(self.optOFW)
			self.grpAvailableFwUpdatesLayout.addWidget(self.lblOFW_Info)
			self.grpAvailableFwUpdatesLayout.addWidget(self.optExternal)
			self.grpAvailableFwUpdatesLayout.addSpacing(3)
			self.grpAvailableFwUpdatesLayout.addItem(self.rowUpdate)
			self.grpAvailableFwUpdates.setLayout(self.grpAvailableFwUpdatesLayout)
			self.layout_device.addWidget(self.grpAvailableFwUpdates)
			# ↑↑↑ Available Firmware Updates

			self.grpStatus = QtWidgets.QGroupBox("")
			self.grpStatusLayout = QtWidgets.QGridLayout()
			self.prgStatus = QtWidgets.QProgressBar()
			self.prgStatus.setMinimum(0)
			self.prgStatus.setMaximum(100)
			self.prgStatus.setValue(0)
			self.lblStatus = QtWidgets.QLabel(__("Ready."))

			self.grpStatusLayout.addWidget(self.prgStatus, 1, 0)
			self.grpStatusLayout.addWidget(self.lblStatus, 2, 0)

			self.grpStatus.setLayout(self.grpStatusLayout)
			self.layout_device.addWidget(self.grpStatus)

			self.grpFooterLayout = QtWidgets.QHBoxLayout()
			self.btnClose = QtWidgets.QPushButton(c__("Button (& = Keyboard Shortcut)", "&Close"))
			self.connect(self.btnClose, QtCore.SIGNAL("clicked()"), lambda: [ self.reject() ])
			self.grpFooterLayout.addStretch()
			self.grpFooterLayout.addWidget(self.btnClose)
			self.layout_device.addItem(self.grpFooterLayout)

			self.layout.addLayout(self.layout_device, 0, 0)
			self.setLayout(self.layout)

			self.ReadDeviceInfo()

		def run(self):
			self.layout.update()
			self.layout.activate()
			screenGeometry = QDesktopWidget().screenGeometry(self)
			x = (screenGeometry.width() - self.width()) / 2
			y = (screenGeometry.height() - self.height()) / 2
			self.move(x, y)
			self.show()

		def hideEvent(self, event):
			if self.DEVICE is None:
				self.APP.ConnectDevice()
			self.APP.activateWindow()

		def reject(self):
			if self.CloseDialog():
				super().reject()

		def CloseDialog(self):
			if self.btnClose.isEnabled() is False:
				text = __("<b>Warning:</b> If you close this window while a firmware update is still running, it might leave the device in an unbootable state.") + "<br><br>" + __("Are you sure you want to close this window?")
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Warning, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
				msgbox.setDefaultButton(QtWidgets.QMessageBox.No)
				answer = msgbox.exec()
				if answer == QtWidgets.QMessageBox.No: return False
			return True

		def ReadDeviceInfo(self):
			self.lblDeviceNameResult.setText(self.DEVICE.GetName())
			self.lblDeviceFWVerResult.setText(self.DEVICE.GetFirmwareVersion(more=True))
			self.lblDevicePCBVerResult.setText(self.DEVICE.GetPCBVersion())

		def ResetAVR(self, delay=0.1):
			port = self.PORT
			try:
				dev = serial.Serial(port, 1000000, timeout=1)
			except serial.SerialException:
				return False
			dev.write(b'0')
			dev.flush()
			time.sleep(0.00125)
			dev.write(struct.pack(">BIBB", 0x2A, 0x37653565, 0x31, 0))
			dev.flush()
			time.sleep(0.00125)
			self.APP.QT_APP.processEvents()
			time.sleep(0.3 + delay)
			dev.reset_input_buffer()
			dev.reset_output_buffer()
			dev.close()
			return True

		def UpdateFirmware(self):
			fw = ""
			path = ""
			if self.optCFW.isChecked():
				fw = self.CFW_VER
				fn = "cfw.hex"
			elif self.optOFW.isChecked():
				fw = self.OFW_VER
				fn = "ofw.hex"
			else:
				path = self.APP.SETTINGS.value("LastDirFirmwareUpdate")
				path = QtWidgets.QFileDialog.getOpenFileName(self, __("Choose GBxCart RW Firmware File"), path, __("Firmware Update") + " (*.hex);;" + __("All Files") + " (*.*)")[0]
				if path == "": return
				temp = re.search(r"^(gbx(?:cart|mas)_rw_.+_pcb_r.+\.hex)$", os.path.basename(path))
				if temp is None:
					msg = __("The expected filename for a valid firmware file is <b>{filename_pattern}</b>. Please visit {url} for the latest official firmware updates.", filename_pattern="gbx*_rw_*_pcb_r*.hex", url="<a href=\"https://www.gbxcart.com/\">https://www.gbxcart.com</a>")
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=msg, standardButtons=QtWidgets.QMessageBox.Ok)
					answer = msgbox.exec()
					return
				self.APP.SETTINGS.setValue("LastDirFirmwareUpdate", os.path.dirname(path))
				fw = "{:s}\n\n".format(path) + __("Please double check that this is a valid firmware file for your GBxCart RW. If it is invalid or an update for a different device, it may render your device unusable.")
				fn = None

			text = __("The following firmware will now be written to your GBxCart RW device:") + "\n- {fw}".format(fw=fw)
			text += "\n\n" + __("Do you want to continue?")
			msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Question, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
			msgbox.setDefaultButton(QtWidgets.QMessageBox.Yes)
			answer = msgbox.exec()
			if answer == QtWidgets.QMessageBox.No: return
			self.btnUpdate.setEnabled(False)
			self.btnClose.setEnabled(False)
			self.grpAvailableFwUpdates.setEnabled(False)

			if path == "":
				with zipfile.ZipFile(self.APP_PATH + os.sep + os.path.join("res", "{:s}".format(self.FW_FILES[self.PCB_VER]))) as archive:
					with archive.open(fn) as f: ihex = f.read().decode("ascii")
			else:
				with open(path, "rb") as f: ihex = f.read().decode("ascii")

			ihex = ihex.splitlines()
			buffer = bytearray()
			for line in ihex:
				keys = ["colon", "raw", "bytecount", "address", "type", "data", "checksum"]
				values = re.search(r"^(\:)((.{2})(.{4})(.{2})(.*))(.{2})$", line)
				if values == None: continue
				values = values.groups()
				data = dict(zip(keys, values))
				for (k, v) in data.items():
					if k in ("bytecount", "type", "checksum"):
						data[k] = struct.unpack("B", bytes.fromhex(v))[0]
					elif k == "address":
						data[k] = struct.unpack("H", bytes.fromhex(v))[0]
					elif k != "colon":
						data[k] = bytes.fromhex(v)

				# Calculate checksum
				chk = 0
				for i in range(0, len(data["raw"])):
					chk += data["raw"][i]
				chk = chk & 0xFF
				chk = (~chk + 1) & 0xFF
				if (chk != data["checksum"]):
					self.SetStatus(__("Firmware checksum error."))
					self.prgStatus.setValue(0)
					self.btnUpdate.setEnabled(True)
					self.btnClose.setEnabled(True)
					self.grpAvailableFwUpdates.setEnabled(True)
					return False

				else:
					buffer += bytearray(data["data"])

			if len(buffer) >= 7168:
				self.SetStatus(__("Firmware file is too large."))
				self.prgStatus.setValue(0)
				self.btnUpdate.setEnabled(True)
				self.btnClose.setEnabled(True)
				self.grpAvailableFwUpdates.setEnabled(True)
				return False

			self.APP.DisconnectDevice()

			while True:
				ret = self.WriteFirmware(buffer, self.SetStatus)
				if ret == 1: return True
				elif ret == 2: return False
				elif ret == 3: continue

		def SetStatus(self, text, enableUI=False, setProgress=None):
			self.lblStatus.setText(__("Status: {text}", text=text))
			if setProgress is not None:
				self.prgStatus.setValue(setProgress)
			if enableUI:
				self.btnUpdate.setEnabled(True)
				self.btnClose.setEnabled(True)
				self.grpAvailableFwUpdates.setEnabled(True)

		def WriteFirmware(self, data, fncSetStatus):
			fw_buffer = data
			port = self.PORT

			delay = 0
			lives = 10
			buffer = bytearray()

			msgWarnBadResponse = __(
				"Failed to update your GBxCart RW {pcb_version} ({fw_version})!\n\n"
				"The firmware update failed as the device is not responding correctly. Please ensure you use a genuine GBxCart RW, re-connect using a different USB cable and try again.\n\n"
				"⚠️ Please note that FlashGBX does not work with the “{flashboy}” series devices.",
				pcb_version=self.PCB_VER,
				fw_version=self.FW_VER,
				flashboy="FLASH BOY"
			)

			fncSetStatus(text=__("Waiting for bootloader..."), setProgress=0)
			if self.ResetAVR(delay) is False:
				fncSetStatus(text=__("Bootloader error."), enableUI=True)
				self.prgStatus.setValue(0)
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME + " – " + __("Firmware Updater for {device_name}", device_name="GBxCart RW " + self.PCB_VER), text=msgWarnBadResponse, standardButtons=QtWidgets.QMessageBox.Ok)
				answer = msgbox.exec()
				return 2

			while True:
				try:
					dev = serial.Serial(port=port, baudrate=9600*4, timeout=1)
				except:
					fncSetStatus(text=__("Device access error."), enableUI=True)
					return 2
				dev.reset_input_buffer()
				dev.reset_output_buffer()
				dev.write(b"@@@")
				dev.flush()
				time.sleep(0.00125)
				buffer = dev.read(0x11)
				if (len(buffer) < 0x11) or (buffer[0:3] != b'TSB'):
					dev.write(b"?")
					dev.flush()
					time.sleep(0.00125)
					dev.close()
					self.APP.QT_APP.processEvents()
					time.sleep(1)
					if len(buffer) != 0x11:
						delay += 0.05
					fncSetStatus(__("Waiting for bootloader... (+{milliseconds}ms)", milliseconds=math.ceil(delay * 1000)))
					if self.ResetAVR(delay) is False:
						fncSetStatus(text=__("Bootloader error."), enableUI=True)
						msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME + " – " + __("Firmware Updater for {device_name}", device_name="GBxCart RW " + self.PCB_VER), text=msgWarnBadResponse, standardButtons=QtWidgets.QMessageBox.Ok)
						answer = msgbox.exec()
						return 2
					lives -= 1
					if lives < 0:
						fncSetStatus(text=__("Bootloader timeout."), enableUI=True)
						msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME + " – " + __("Firmware Updater for {device_name}", device_name="GBxCart RW " + self.PCB_VER), text=msgWarnBadResponse, standardButtons=QtWidgets.QMessageBox.Ok)
						answer = msgbox.exec()
						return 2
					continue
				break

			fncSetStatus(__("Reading bootloader information..."))
			info = {}
			keys = ["magic", "tsb_version", "tsb_status", "signature", "page_size", "flash_size", "eeprom_size", "unknown", "avr_jmp_identifier"]
			values = struct.unpack("<3sHB3sBHHBB", bytearray(buffer[:-1]))
			info = dict(zip(keys, values))
			info["page_size"] *= 2
			info["flash_size"] *= 2
			info["eeprom_size"] += 1
			if info["avr_jmp_identifier"] == 0x00:
				info["jmp_mode"] = "relative"
				info["device_type"] = "attiny"
			elif info["avr_jmp_identifier"] == 0x0C:
				info["jmp_mode"] = "absolute"
				info["device_type"] = "attiny"
			elif info["avr_jmp_identifier"] == 0xAA:
				info["jmp_mode"] = "relative"
				info["device_type"] = "atmega"

			if info["page_size"] != 64 or info["flash_size"] != 7616 or info["eeprom_size"] != 512 or info["jmp_mode"] != "relative" or info["device_type"] != "atmega" or info["signature"] != b'\x1E\x93\x06':
				fncSetStatus(text="Wrong device detected.", enableUI=True)
				return 2

			if (info["tsb_version"] < 32768):
				info["tsb_version"] = int((info["tsb_version"] & 31) + ((info["tsb_version"] & 480) / 32) * 100 + ((info["tsb_version"] & 65024 ) / 512) * 10000 + 20000000)
			else:
				fncSetStatus(text="Wrong device detected.", enableUI=True)
				return 2

			#################

			# Read user data
			fncSetStatus(__("Reading user data..."))
			dev.write(b"c")
			user_data = bytearray(dev.read(0x41))
			info["tsb_timeout"] = user_data[2]

			# Change timeout to 6s
			fncSetStatus(__("Writing user data..."))
			user_data[2] = 254
			dev.write(b"C")
			dev.read(1)
			dev.write(b"!")
			dev.write(user_data)
			dev.flush()
			time.sleep(0.00125)
			dev.read(0x41)

			# Write firmware
			fncSetStatus(__("Updating firmware... Do not unplug the device!"))
			iterations = math.ceil(len(fw_buffer) / 0x40)
			if len(fw_buffer) < iterations * 0x40:
				fw_buffer = fw_buffer + bytearray([0xFF] * ((iterations * 0x40) - len(fw_buffer)))

			lives = 10
			dev.write(b"F")
			dev.flush()
			time.sleep(0.00125)
			ret = dev.read(1)
			while ret != b"?":
				dev.write(b"F")
				dev.flush()
				time.sleep(0.00125)
				ret = dev.read(1)
				lives -= 1
				if lives == 0:
					dev.write(b"?")
					dev.flush()
					time.sleep(0.00125)
					dev.close()
					fncSetStatus(text=__("Protocol Error. Please try again."), enableUI=True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text="The firmware update was not successful (Protocol Error). Do you want to try again?\n\nIf it doesn’t work even after multiple retries, please use the insideGadgets standalone firmware updater instead.", standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, defaultButton=QtWidgets.QMessageBox.Yes)
					answer = msgbox.exec()
					if answer == QtWidgets.QMessageBox.Yes:
						time.sleep(1)
						return 3
					return 2

			for i in range(0, iterations):
				self.APP.QT_APP.processEvents()
				dev.write(b"!")
				dev.write(fw_buffer[i*0x40:i*0x40+0x40])
				fncSetStatus(text=__("Updating firmware... Do not unplug the device!"), setProgress=(i*0x40+0x40) / len(fw_buffer) * 100)
				ret = dev.read(1)
				if (ret != b"?"):
					dev.write(b"?")
					dev.flush()
					time.sleep(0.00125)
					dev.close()
					fncSetStatus(text=__("Write Error ({error_value}). Please try again.", error_value=str(ret)), enableUI=True)
					msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=__("The firmware update was not successful (Write Error, {error_value}). Do you want to try again?\n\nIf it doesn’t work even after multiple retries, please use the insideGadgets standalone firmware updater instead.", error_value=str(ret)), standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, defaultButton=QtWidgets.QMessageBox.Yes)
					answer = msgbox.exec()
					if answer == QtWidgets.QMessageBox.Yes:
						time.sleep(1)
						return 3
					return 2
			dev.write(b"?")
			dev.flush()
			time.sleep(0.00125)
			dev.read(1)

			# verify flash
			fncSetStatus(__("Verifying update..."))
			buffer2 = bytearray()
			dev.write(b"f")
			dev.flush()
			time.sleep(0.00125)
			for i in range(0, 0x1DC0, 0x40):
				self.APP.QT_APP.processEvents()
				dev.write(b"!")
				dev.flush()
				time.sleep(0.00125)
				while dev.in_waiting == 0: time.sleep(0.01)
				ret = bytearray(dev.read(0x40))
				buffer2 += ret
				self.prgStatus.setValue(len(buffer2) / 0x1DC0 * 100)
			dev.read(1)

			buffer2 = buffer2[:len(fw_buffer)]

			if fw_buffer == buffer2:
				fncSetStatus(__("Verification OK."))
				self.APP.QT_APP.processEvents()
				time.sleep(0.2)
			else:
				fncSetStatus(text=__("Verification Error."), enableUI=True)
				dev.write(b"?")
				dev.flush()
				time.sleep(0.00125)
				dev.close()
				msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Critical, windowTitle=AppInfo.NAME, text=__("The firmware update was not successful (Verification Error). Do you want to try again?\n\nIf it doesn’t work even after multiple retries, please use the insideGadgets standalone firmware updater instead."), standardButtons=QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No, defaultButton=QtWidgets.QMessageBox.Yes)
				answer = msgbox.exec()
				if answer == QtWidgets.QMessageBox.Yes:
					time.sleep(1)
					return 3
				return 2

			# Change timeout to 1s
			fncSetStatus(__("Writing user data..."))
			user_data[2] = 42
			dev.write(b"C")
			dev.flush()
			time.sleep(0.00125)
			ret = dev.read(1)
			while ret != b"?":
				dev.write(b"C")
				dev.flush()
				time.sleep(0.00125)
				ret = dev.read(1)
				lives -= 1
				if lives == 0:
					dev.write(b"?")
					dev.flush()
					time.sleep(0.00125)
					dev.close()
					fncSetStatus(text=__("User data update error. Please try again."), enableUI=True)
					return 2
			dev.write(b"!")
			dev.write(user_data)
			dev.flush()
			time.sleep(0.00125)
			dev.read(0x41)

			# Restart
			self.APP.QT_APP.processEvents()
			time.sleep(0.1)
			fncSetStatus(__("Restarting the device..."))
			dev.write(b"?")
			dev.flush()
			time.sleep(0.00125)
			dev.close()
			self.APP.QT_APP.processEvents()
			time.sleep(0.8)
			fncSetStatus(__("Done!"))
			self.APP.QT_APP.processEvents()
			time.sleep(0.2)
			self.DEVICE = None
			self.btnUpdate.setEnabled(True)
			self.btnClose.setEnabled(True)
			self.grpAvailableFwUpdates.setEnabled(True)
			text = __("The firmware update is complete!")
			msgbox = QtWidgets.QMessageBox(parent=self, icon=QtWidgets.QMessageBox.Information, windowTitle=AppInfo.NAME, text=text, standardButtons=QtWidgets.QMessageBox.Ok)
			answer = msgbox.exec()
			self.reject()
			return 1
except ImportError:
	pass



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

import sys, os, glob, re, json, zlib, argparse, zipfile, traceback, platform, datetime, copy, time
from .i18n import __, c__, init_language
from .app import AppInfo, AppContext, HW_DEVICES
from .IniSettings import IniSettings
from .CartridgeTypes import RomSizes, AgbSaveTypes, DmgSaveTypes
from .Flashcart import empty_flashcarts_map
from .Logging import ANSI
from .PocketCamera import PocketCamera

STATIC_ACTIONS = ["info", "backup-rom", "flash-rom", "backup-save",
                  "restore-save", "erase-save", "gbcamera-extract",
                  "interactive", "debug-test-save"]
FWUPDATE_ACTIONS = []
for _d in HW_DEVICES:
	try:
		_dev = _d.GbxDevice()
		if _dev.SupportsFirmwareUpdates():
			_action = _dev.FirmwareUpdateAction()
			if _action is not None:
				FWUPDATE_ACTIONS.append(_action)
	except Exception:
		pass
ALL_ACTIONS = STATIC_ACTIONS + FWUPDATE_ACTIONS

def ReadConfigFiles(args):
	reset = args['argparsed'].reset
	settings = IniSettings(path=args["config_path"] + os.sep + "settings.ini")
	config_version = settings.value("ConfigVersion")
	if not os.path.exists(args["config_path"]): os.makedirs(args["config_path"])
	fc_files = glob.glob("{:s}{:s}fc_*.txt".format(glob.escape(args["config_path"]), os.sep))
	if config_version is not None and len(fc_files) == 0:
		print(__("No flashcart profile files found in {config_path}. Resetting configuration...", config_path=args["config_path"]))
		settings.clear()
		os.rename(args["config_path"] + os.sep + "settings.ini", args["config_path"] + os.sep + "settings.ini_" + datetime.datetime.now().strftime("%Y%m%d%H%M%S") + ".bak")
		config_version = False # extracts the config.zip again
	elif reset:
		settings.clear()
		print(__("All configuration has been reset."))

	if config_version != AppInfo.VERSION:
		settings.setValue("UpdateCheck", None, quiet=True)
	settings.setValue("ConfigVersion", AppInfo.VERSION, quiet=True)
	return (config_version, fc_files)

def LoadConfig(args):
	app_path = args['app_path']
	config_path = args['config_path']
	ret = []
	flashcarts = empty_flashcarts_map()

	# Settings and Config
	(config_version, fc_files) = ReadConfigFiles(args=args)
	if config_version != AppInfo.VERSION:
		# Rename old files that have since been replaced/renamed/merged
		deprecated_files = [ "fc_AGB_TEST.txt", "fc_DMG_TEST.txt", "fc_AGB_Nintendo_E201850.txt", "fc_AGB_Nintendo_E201868.txt", "config.ini", "fc_DMG_MX29LV320ABTC.txt", "fc_DMG_iG_4MB_MBC3_RTC.txt", "fc_AGB_Flash2Advance.txt", "fc_AGB_MX29LV640_AUDIO.txt", "fc_AGB_M36L0R7050T.txt", "fc_AGB_M36L0R8060B.txt", "fc_AGB_M36L0R8060T.txt", "fc_AGB_iG_32MB_S29GL512N.txt", "fc_DMG_SST39SF010_MBC1_AUDIO.txt", "fc_DMG_SST39SF040_MBC5_AUDIO.txt", "fc_DMG_AM29F010_MBC1_AUDIO.txt", "fc_DMG_AM29F040_MBC1_AUDIO.txt", "fc_DMG_AM29F040_MBC1_WR.txt", "fc_DMG_AM29F080_MBC1_AUDIO.txt", "fc_DMG_AM29F080_MBC1_WR.txt", "fc_DMG_SST39SF040_MBC1_AUDIO.txt", "fc_DMG_SST39SF020_MBC1_AUDIO.txt", "fc_DMG_29LV016T.txt", "fc_DMG_Retrostage.txt" ]
		for file in deprecated_files:
			if os.path.exists(config_path + os.sep + file):
				os.rename(config_path + os.sep + file, config_path + os.sep + file + "_" + datetime.datetime.now().strftime("%Y%m%d%H%M%S") + ".bak")

		rf_list = ""
		if os.path.exists(app_path + os.sep + os.path.join("res", "config.zip")):
			try:
				with zipfile.ZipFile(app_path + os.sep + os.path.join("res", "config.zip")) as zip:
					for zfile in zip.namelist():
						if os.path.exists(config_path + os.sep + zfile):
							zfile_crc = zip.getinfo(zfile).CRC
							with open(config_path + os.sep + zfile, "rb") as ofile: buffer = ofile.read()
							ofile_crc = zlib.crc32(buffer) & 0xFFFFFFFF
							if zfile_crc == ofile_crc: continue
							os.rename(config_path + os.sep + zfile, config_path + os.sep + zfile + "_" + datetime.datetime.now().strftime("%Y%m%d%H%M%S") + ".bak")
							rf_list += zfile + "\n"
						zip.extract(zfile, config_path + os.sep)
			except zipfile.BadZipFile:
				print(__("Warning: config.zip is corrupted and could not be read."))

			if rf_list != "":
				ret.append([1, __("The application was recently updated and some flashcart profile files have been updated as well. You will find backup copies of them in your configuration directory.") + "\n\n" + __("Updated files:") + "\n" + rf_list[:-1]])
			fc_files = glob.glob("{0:s}{1}fc_*.txt".format(glob.escape(config_path), os.sep))
		else:
			print(__("Warning: {config_zip_file} not found. This is required to load new flashcart profile configurations after updating.", config_zip_file=app_path + os.sep + os.path.join("res", "config.zip")))

	# Read flash cart types
	for file in fc_files:
		if os.path.exists(file):
			with open(file, encoding='utf-8') as f:
				data = f.read()
				specs_int = re.sub("(0x[0-9A-F]+)", lambda m: str(int(m.group(1), 16)), data) # hex numbers to int numbers, otherwise not valid json
				try:
					specs = json.loads(specs_int)
				except:
					ret.append([2, "The flashchip type file “{:s}” could not be parsed and needs to be fixed before it can be used.".format(os.path.basename(file))])
					continue
				if "names" not in specs: continue
				for name in specs["names"]:
					if not specs["type"] in flashcarts: continue # only DMG and AGB are supported right now
					temp = copy.deepcopy(specs)
					temp["names"] = [name]
					flashcarts[specs["type"]][name] = temp

	return { "flashcarts":flashcarts, "config_ret":ret }

class ArgParseCustomFormatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter): pass
def main(portableMode=False):
	if platform.system() == "Windows":
		os.system("color")
	elif platform.system() == "Darwin":
		macos_version = tuple(map(int, platform.mac_ver()[0].split('.')))
		try:
			macos_version = tuple(map(int, platform.mac_ver()[0].split('.')))
		except (ValueError, IndexError):
			macos_version = (0, 0)

		# macOS above Big Sur don't need a compat layer fix in the environment
		if macos_version < (12, 0):
			os.environ['QT_MAC_WANTS_LAYER'] = '1'

	AppContext.LAUNCH_TIMESTAMP = time.time()

	if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
		app_path = os.path.dirname(sys.executable)
	else:
		app_path = os.path.dirname(os.path.abspath(__file__))

	try:
		from .pyside import QtCore
		cp = { "subdir":os.path.join(app_path, "config"), "appdata":os.path.join(QtCore.QDir.toNativeSeparators(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.AppConfigLocation)), "FlashGBX") }
	except:
		cp = { "subdir":os.path.join(app_path, "config"), "appdata":os.path.join(os.path.expanduser('~'), "FlashGBX") }

	if portableMode:
		cfgdir_default = "subdir"
	else:
		cfgdir_default = "appdata"

	config_path = None
	language_choice = None
	for i, arg in enumerate(sys.argv):
		if arg == "--cfgdir" and i + 1 < len(sys.argv):
			cfgdir_choice = sys.argv[i + 1].lower()
			if cfgdir_choice in cp:
				config_path = cp[cfgdir_choice]
		elif arg == "--language" and i + 1 < len(sys.argv):
			language_choice = sys.argv[i + 1].lower()

	if config_path is None:
		if cfgdir_default in cp:
			config_path = cp[cfgdir_default]
		else:
			config_path = cp["subdir"]

	init_language(config_path, override=language_choice)

	print("FlashGBX {version}\n© 2020–{year} Lesserkuma".format(version=AppInfo.VERSION, year=time.strftime("%Y")))
	print("https://github.com/Lesserkuma/FlashGBX")

	examples = "\n" + __("Examples") + ":\n" + \
	"  " + __("Backup the ROM of a Game Boy Advance cartridge") + ":\n\tFlashGBX --mode agb --action backup-rom\n\n" + \
	"  " + __("Backup Save Data from a Game Boy cartridge") + ":\n\tFlashGBX --mode dmg --action backup-save\n\n" + \
	"  " + __("Write a Game Boy Advance ROM relying on auto-detecting the flash cartridge") + ":\n\tFlashGBX --mode agb --action flash-rom ROM.gba\n\n" + \
	"  " + __("Extract Game Boy Camera pictures as .png files from a save data file") + ":\n\tFlashGBX --mode dmg --action gbcamera-extract --gbcamera-outfile-format png GAMEBOYCAMERA.sav\n\n" + \
	"  " + __("Backup a {gb_memory_cartridge} ROM including its hidden sector .map file", gb_memory_cartridge="NP GB-Memory Cartridge") + ":\n\tFlashGBX --mode dmg --action backup-rom --dmg-mbc 0x105\n\n"

	parser = argparse.ArgumentParser(formatter_class=ArgParseCustomFormatter, epilog=examples)
	try:
		# pylint: disable=protected-access
		parser._action_groups[1].title = c__("Command Line Arguments Category", "General arguments")
	except:
		pass
	parser.add_argument("--cli", help=c__("Command Line Help", "force command line interface mode"), action="store_true")
	parser.add_argument("--reset", help=c__("Command Line Help", "clears all settings such as last used directory information"), action="store_true")
	parser.add_argument("--debug", help=c__("Command Line Help", "enable debug messages used for development"), action="store_true")
	parser.add_argument("--language", action="store", help=c__("Command Line Help", "sets the language of the program (e.g. “auto”, “en”, “de”, ...)"))

	parser.add_argument_group('')
	ap_config = parser.add_argument_group(c__("Command Line Arguments Category", "Configuration arguments"))
	if "appdata" in cp: ap_config.add_argument("--cfgdir", choices=["appdata", "subdir"], type=str.lower, default=cfgdir_default, help=c__("Configuration Help", "sets the config directory to either the OS-provided local app config directory ({appdata}), or a subdirectory of this application ({subdir})", appdata=cp["appdata"], subdir=cp["subdir"]))

	ap_cli1 = parser.add_argument_group(c__("Command Line Arguments Category", "Main command line interface arguments"))
	ap_cli1.add_argument("--mode", choices=["dmg", "agb"], type=str.lower, default=None, help=c__("Command Line Help", "set platform to “dmg” (Game Boy) or “agb” (Game Boy Advance)"))
	ap_cli1.add_argument("--action", choices=ALL_ACTIONS, type=str.lower, default=None, help=c__("Command Line Help", "select program action"))
	ap_cli1.add_argument("--overwrite", action="store_true", help=c__("Command Line Help", "overwrite without asking if target file already exists"))
	ap_cli1.add_argument("path", nargs="?", default="auto", help=c__("Command Line Help", "target or source file path (optional when reading, required when writing)"))

	ap_cli2 = parser.add_argument_group(c__("Command Line Arguments Category", "Optional command line interface arguments"))
	ap_cli2.add_argument("--dmg-romsize", choices=RomSizes.GetCLINames(mode="DMG"), type=str.lower, default="auto", help=c__("Command Line Help", "set size of Game Boy cartridge ROM data"))
	ap_cli2.add_argument("--dmg-mbc", type=str.lower, default="auto", help=c__("Command Line Help", "set mapper type of Game Boy cartridge"))
	ap_cli2.add_argument("--dmg-savetype", choices=DmgSaveTypes.GetCLINames(), type=str.lower, default="auto", help=c__("Command Line Help", "set type of Game Boy cartridge save data"))
	ap_cli2.add_argument("--agb-romsize", choices=RomSizes.GetCLINames(mode="AGB"), type=str.lower, default="auto", help=c__("Command Line Help", "set size of Game Boy Advance cartridge ROM data"))
	ap_cli2.add_argument("--agb-savetype", choices=AgbSaveTypes.GetCLINames(), type=str.lower, default="auto", help=c__("Command Line Help", "set type of Game Boy Advance cartridge save data"))
	ap_cli2.add_argument("--bl-offset", type=str, default="auto", help=c__("Command Line Help", "Location of Batteryless SRAM data in ROM (e.g. 0xFC0000); only with “{dmg_savetype_batteryless}” or “{agb_savetype_batteryless}”)", dmg_savetype_batteryless="--dmg-savetype batteryless", agb_savetype_batteryless="--agb-savetype batteryless"))
	ap_cli2.add_argument("--bl-size", type=str, default="auto", help=c__("Command Line Help", "Size of Batteryless SRAM data in ROM (e.g. 0x10000)"))
	ap_cli2.add_argument("--bl-layout", choices=["auto", "0", "1", "2"], type=str.lower, default="auto", help=c__("Command Line Help", "Bank layout of Batteryless SRAM data for DMG mode: 0=continuous, 1=first half of ROM bank, 2=second half"))
	ap_cli2.add_argument("--store-rtc", action="store_true", default=False, help=c__("Command Line Help", "store RTC register values if supported"))
	ap_cli2.add_argument("--keep-calibration", action="store_true", default=True, help=c__("Command Line Help", "keep existing calibration data of the e-Reader when writing save data"))
	ap_cli2.add_argument("--ignore-bad-header", action="store_true", help=c__("Command Line Help", "don’t stop if invalid data found in cartridge header data"))
	ap_cli2.add_argument("--flashcart-type", type=str, default="autodetect", help=c__("Command Line Help", "name of flash cart profile; see .txt files in config directory"))
	ap_cli2.add_argument("--prefer-chip-erase", action="store_true", help=c__("Command Line Help", "prefer full chip erase over sector erase when both available"))
	ap_cli2.add_argument("--force-5v", action="store_true", help=c__("Command Line Help", "force 5V when writing Game Boy flash cartridges"))
	ap_cli2.add_argument("--no-verify-write", action="store_true", help=c__("Command Line Help", "do not verify written data"))
	ap_cli2.add_argument("--generate-dump-report", action="store_true", help=c__("Command Line Help", "generate a dump report for a ROM backup"))
	ap_cli2.add_argument("--save-filename-add-datetime", action="store_true", help=c__("Command Line Help", "adds a timestamp to the file name of save data backups"))
	ap_cli2.add_argument("--gbcamera-palette", choices=PocketCamera.PALETTE_NAMES, type=str.lower, default="grayscale", help=c__("Command Line Help", "sets the color palette of pictures extracted from Game Boy Camera saves"))
	ap_cli2.add_argument("--gbcamera-outfile-format", choices=PocketCamera.OUTPUT_FORMATS, type=str.lower, default="png", help=c__("Command Line Help", "sets the file format of saved pictures extracted from Game Boy Camera saves"))
	ap_cli2.add_argument("--gbcamera-extract", action="store_true", default=False, help=c__("Command Line Help", "automatically extract Game Boy Camera pictures after backing up save data"))
	ap_cli2.add_argument("--device-port", help=c__("Command Line Help", "override device port"), default=None)
	ap_cli2.add_argument("--device-limit-baudrate", action="store_true", help=c__("Command Line Help", "limit connection to a slower baud rate"))
	ap_cli2.add_argument("--compare-sectors", action="store_true", help=c__("Command Line Help", "compare sectors and only write those that differ when writing a ROM (only for flash carts that support this feature)"), default=True)
	ap_cli2.add_argument("--wait", action="store_true", help=c__("Command Line Help", "wait for key press after the program has ended"))
	args = None
	try:
		args, _ = parser.parse_known_args()
	except SystemExit:
		if args is None or "--help" in sys.argv:
			input("\n\n" + __("Press ENTER to exit.") + "\n")
			return 0

	if "appdata" in cp and hasattr(args, 'cfgdir'):
		parsed_config_path = cp[args.cfgdir]
		if parsed_config_path != config_path:
			config_path = parsed_config_path

	if args.mode is not None or args.action is not None:
		args.cli = True

	if args.debug == True:
		AppContext.DEBUG = True

	args = {"app_path":app_path, "config_path":config_path, "argparsed":args}
	while True:
		try:
			if not os.path.exists(config_path):
				os.mkdir(config_path)
			tf = "{:s}/settings.ini".format(config_path)
			f = open(tf, "ab")
			f.close()
			break
		except PermissionError:
			print("\n" + ANSI.RED + __("Error: This program has no permission to use the configuration directory “{config_path}”!", config_path=config_path) + ANSI.RESET)
			if "appdata" in cp and args["argparsed"].cfgdir == "subdir":
				answer = input(__("Use directory “{appdata_folder}” instead?", appdata_folder=cp["appdata"]) + " [y/N] ").strip().lower()
				if answer != "y":
					return
				config_path = cp["appdata"]
				args["config_path"]  = config_path
				continue
			else:
				input("")
			if args["argparsed"].wait: input("\n\n" + __("Press ENTER to exit.") + "\n")
			return

	args.update(LoadConfig(args))

	app = None
	exc = None
	retval = -1
	if not args["argparsed"].cli:
		try:
			from . import FlashGBX_GUI
			app = FlashGBX_GUI.FlashGBX_GUI(args)
		except ModuleNotFoundError:
			exc = traceback.format_exc()
			app = None
		except:
			exc = traceback.format_exc()
			app = None

		if app is None:
			from . import FlashGBX_CLI
			if args["argparsed"].action is None:
				parser.print_help()
				print("\n\n{:s}" + __("Note: GUI mode couldn’t be launched, but the application can be run in CLI mode.") + "\n      " + __("Optional command line switches are explained above.") + "{:s}\n".format(ANSI.RED, ANSI.RESET))
				if exc is not None: print(ANSI.YELLOW + str(exc) + ANSI.RESET)

			print(__("Falling back to CLI mode.") + "\n")
			app = FlashGBX_CLI.FlashGBX_CLI(args)
			try:
				retval = app.run()
			except KeyboardInterrupt:
				print("\n\n" + __("Program stopped."))
			if args["argparsed"].wait: input("\n" + __("Press ENTER to exit.") + "\n")
			sys.exit(retval)

		app.run()

	else:
		from . import FlashGBX_CLI
		print("\n" + __("Now running in CLI mode."))
		app = FlashGBX_CLI.FlashGBX_CLI(args)
		try:
			retval = app.run()
		except KeyboardInterrupt:
			print("\n\n" + __("Program stopped."))
		if args["argparsed"].wait: input("\n" + __("Press ENTER to exit.") + "\n")
		sys.exit(retval)



# -*- coding: utf-8 -*-
# FlashGBX
# Author: Lesserkuma (github.com/Lesserkuma)

from .i18n import __, c__

class RomSizes:
	ROM_SIZES = [2**i for i in range(15, 30)] # 32 KiB to 512 MiB
	ROM_SIZES_DMG = [2**i for i in range(15, 28)] # 32 KiB to 128 MiB

	def __init__(self, size=None, index=None):
		if index is not None:
			self._size = self.GetSize(index)
		elif size is not None:
			self._size = size
		else:
			self._size = None

	def __contains__(self, size):
		return size in self.ROM_SIZES

	def GetString(self, size=None, index=None, localized=True):
		if index is not None:
			if not (0 <= index < len(self.ROM_SIZES)): return ""
			size = self.ROM_SIZES[index]
		elif size is None:
			size = self._size

		if size is None:
			size = 0

		bytes = __(" Bytes") if localized else " Bytes"
		kib = __(" KiB") if localized else " KiB"
		mib = __(" MiB") if localized else " MiB"

		if size >= 1024 * 1024:
			value = size // (1024 * 1024)
			return f"{value}{mib}"
		elif size >= 1024:
			value = size // 1024
			return f"{value}{kib}"
		else:
			return f"{size}{bytes}"

	def GetNextLarger(self, size):
		for valid_size in self.ROM_SIZES:
			if valid_size >= size:
				return valid_size
		return None

	def GetSize(self, index):
		if index < len(self.ROM_SIZES):
			return self.ROM_SIZES[index]
		return None

	def GetIndex(self, size):
		if size in self.ROM_SIZES:
			return self.ROM_SIZES.index(size)
		return None

	def GetStringList(self, mode="AGB"):
		if mode == "DMG":
			return [self.GetString(index=index) for index in range(len(self.ROM_SIZES_DMG))]
		elif mode == "AGB":
			return [self.GetString(index=index) for index in range(len(self.ROM_SIZES))]
		return []

	def GetNumberOfTypes(self, mode="AGB"):
		if mode == "DMG":
			return len(self.ROM_SIZES_DMG)
		return len(self.ROM_SIZES)

	@classmethod
	def _SizeToCLIName(cls, size):
		if size >= 1024 * 1024:
			return f"{size // (1024 * 1024)}mb"
		return f"{size // 1024}kb"

	@classmethod
	def GetCLINames(cls, mode="AGB", include_auto=True):
		sizes = cls.ROM_SIZES_DMG if mode == "DMG" else cls.ROM_SIZES
		names = [cls._SizeToCLIName(s) for s in sizes]
		return (["auto"] + names) if include_auto else names

	@classmethod
	def GetSizeFromCLIName(cls, name, mode="AGB"):
		if name == "auto":
			return None
		sizes = cls.ROM_SIZES_DMG if mode == "DMG" else cls.ROM_SIZES
		for s in sizes:
			if cls._SizeToCLIName(s) == name:
				return s
		return None


class AgbSaveTypes:
	SAVE_TYPES = [
		(0, ""),
		(512, "4K EEPROM"),
		(8192, "64K EEPROM"),
		(32768, "256K SRAM/FRAM"),
		(65536, "512K FLASH"),
		(131072, "1M FLASH"),
		(1048576, "8M DACS"),
		(65536, "Unlicensed 512K SRAM"),
		(131072, "Unlicensed 1M SRAM"),
		(0, "Unlicensed Batteryless SRAM")
	]

	# CLI shorthand for each SAVE_TYPES entry. None means the entry has no CLI alias.
	CLI_NAMES = [
		None,           # 0
		"eeprom4k",     # 1
		"eeprom64k",    # 2
		"sram256k",     # 3
		"flash512k",    # 4
		"flash1m",      # 5
		"dacs8m",       # 6
		"sram512k",     # 7
		"sram1m",       # 8
		"batteryless",  # 9
	]

	AGB_FLASH_SAVE_CHIPS = {
		0xBFD4: ("SST 39VF512", 0x10000),
		0x1F3D: ("Atmel AT29LV512", 0x10000),
		0xC21C: ("Macronix MX29L512", 0x10000),
		0x321B: ("Panasonic MN63F805MNP", 0x10000),
		0xC209: ("Macronix MX29L010", 0x20000),
		0x6213: ("SANYO LE26FV10N1TS", 0x20000),
		0xBF4B: ("Unlicensed SST25VF064C", 0x20000),
		0xBF5B: ("Unlicensed SST49LF080A", 0x20000),
		0xBF6D: ("Unlicensed SST39VF6401B", 0x20000),
		0xFFFF: ("Unlicensed 0xFFFF", 0x20000)
	}

	def __init__(self, index=None):
		self._index = index

	def __contains__(self, size):
		return size in [s[0] for s in self.SAVE_TYPES]

	def GetName(self, index=None):
		if index is None:
			index = self._index
		if index is not None and 0 <= index < len(self.SAVE_TYPES):
			return self.SAVE_TYPES[index][1]
		return c__("Game Data", "Unknown")

	def GetSize(self, index=None):
		if index is None:
			index = self._index
		if index is not None and 0 <= index < len(self.SAVE_TYPES):
			return self.SAVE_TYPES[index][0]
		return None

	def GetString(self, index=None, localized=True):
		if index is None:
			index = self._index

		if index is None or index < 0 or index >= len(self.SAVE_TYPES):
			return c__("Game Data", "Unknown")

		name = self.GetName(index)
		if name == "":
			name = c__("Save Type", "None") if localized else "None"
		bytes_val = self.GetSize(index)

		if bytes_val == 0:
			return name

		bytes = __(" Bytes") if localized else " Bytes"
		kib = __(" KiB") if localized else " KiB"
		mib = __(" MiB") if localized else " MiB"

		if bytes_val >= 1024 * 1024:
			return f"{name} ({bytes_val >> 20}{mib})"
		elif bytes_val >= 1024:
			return f"{name} ({bytes_val >> 10}{kib})"
		else:
			return f"{name} ({bytes_val}{bytes})"

	def GetIndexFromSize(self, size):
		for idx, (bytes_val, _) in enumerate(self.SAVE_TYPES):
			if bytes_val == size:
				return idx
		return None

	def GetStringFromSaveLib(self, savelib_string, localized=True):
		if not savelib_string or savelib_string == "N/A":
			return __("None") if localized else "None"
		elif "SRAM_F_" in savelib_string:
			return "256K SRAM/FRAM ({:s})".format(savelib_string)
		elif "SRAM_" in savelib_string:
			return "256K SRAM ({:s})".format(savelib_string)
		elif "EEPROM_V" in savelib_string:
			return "4K or 64K EEPROM ({:s})".format(savelib_string)
		elif "FLASH_V" in savelib_string or "FLASH512_V" in savelib_string:
			return "512K FLASH ({:s})".format(savelib_string)
		elif "FLASH1M_V" in savelib_string:
			return "1M FLASH ({:s})".format(savelib_string)
		elif "AGB_8MDACS_DL_V" in savelib_string:
			return "8M DACS ({:s})".format(savelib_string)
		else:
			return c__("Save Type", "Unknown") + " ({:s})".format(savelib_string) if localized else "Unknown ({:s})".format(savelib_string)

	def GetStringList(self):
		return [self.GetString(index) for index in range(len(self.SAVE_TYPES))]

	def GetNumberOfTypes(self):
		return len(self.SAVE_TYPES)

	def GetFlashChipName(self, chip_index):
		if not self.IsValidFlashChipIndex(chip_index): return "Unknown"
		return self.AGB_FLASH_SAVE_CHIPS[chip_index][0]

	def GetFlashChipSize(self, chip_index):
		if not self.IsValidFlashChipIndex(chip_index): return 0
		return self.AGB_FLASH_SAVE_CHIPS[chip_index][1]

	def IsValidFlashChipIndex(self, chip_index):
		return chip_index in self.AGB_FLASH_SAVE_CHIPS.keys()

	@classmethod
	def GetCLINames(cls, include_auto=True):
		out = ["auto"] if include_auto else []
		out.extend(name for name in cls.CLI_NAMES if name is not None)
		return out

	@classmethod
	def GetIndexFromCLIName(cls, name):
		if name == "auto":
			return None
		if name in cls.CLI_NAMES:
			return cls.CLI_NAMES.index(name)
		return None

class DmgSaveTypes:
	RAM_TYPES = [
		(0x00, 0, __("None")),
		(0x100, 0x200, "4K SRAM"),
		(0x01, 0x800, "16K SRAM"),
		(0x02, 0x2000, "64K SRAM"),
		(0x03, 0x8000, "256K SRAM"),
		(0x05, 0x10000, "512K SRAM"),
		(0x04, 0x20000, "1M SRAM"),
		(0x104, 0x108000, "MBC6 SRAM+FLASH"),
		(0x101, 0x100, "MBC7 2K EEPROM"),
		(0x102, 0x200, "MBC7 4K EEPROM"),
		(0x103, 0x20, "TAMA5 EEPROM"),
		(0x201, 0x80000, "Unlicensed 4M SRAM"),
		(0x203, 0x20000, "Unlicensed 1M EEPROM"),
		(0x204, 0x100000, "Unlicensed Photo! Directory"),
		(0x205, 0, "Unlicensed Batteryless SRAM"),
	]

	# CLI shorthand → mbc id. The 0x00 "None" entry deliberately has no shorthand
	# (argparse "auto" covers that case).
	CLI_NAMES = {
		0x100: "4k",
		0x01:  "16k",
		0x02:  "64k",
		0x03:  "256k",
		0x05:  "512k",
		0x04:  "1m",
		0x104: "mbc6",
		0x101: "mbc7_2k",
		0x102: "mbc7_4k",
		0x103: "tama5",
		0x201: "sram4m",
		0x203: "eeprom1m",
		0x204: "photo",
		0x205: "batteryless",
	}

	def __init__(self, size=None, index=None, mbc=None):
		if index is not None:
			self._entry = self.RAM_TYPES[index] if 0 <= index < len(self.RAM_TYPES) else None
		elif mbc is not None:
			self._entry = self._FindByMbc(mbc)
		elif size is not None:
			self._entry = self._FindBySize(size)
		else:
			self._entry = None

	def _FindByMbc(self, mbc):
		for entry in self.RAM_TYPES:
			if entry[0] == mbc:
				return entry
		return None

	def _FindBySize(self, size):
		for entry in self.RAM_TYPES:
			if entry[1] == size:
				return entry
		return None

	def GetName(self):
		return self._entry[2] if self._entry else __("Unknown Save Type")

	def GetSize(self):
		return self._entry[1] if self._entry else 0

	def GetMbc(self):
		return self._entry[0] if self._entry else None

	def GetIndex(self):
		if self._entry is None:
			return None
		for index, entry in enumerate(self.RAM_TYPES):
			if entry == self._entry:
				return index
		return None

	def GetString(self, index=None, localized=True):
		entry = self.RAM_TYPES[index] if (index is not None and 0 <= index < len(self.RAM_TYPES)) else self._entry
		name = entry[2] if entry else __("Unknown Save Type")
		size = entry[1] if entry else 0

		if size == 0:
			return name

		bytes = __(" Bytes") if localized else " Bytes"
		kib = __(" KiB") if localized else " KiB"
		mib = __(" MiB") if localized else " MiB"

		if size == 0x108000:
			return f"{name} ({size / (1024 * 1024):.2f}{mib})"
		elif size >= 1024 * 1024:
			return f"{name} ({size // (1024 * 1024)}{mib})"
		elif size >= 1024:
			return f"{name} ({size // 1024}{kib})"
		else:
			return f"{name} ({size}{bytes})"

	def GetStringList(self):
		return [self.GetString(index=index) for index in range(len(self.RAM_TYPES))]

	def GetNumberOfTypes(self):
		return len(self.RAM_TYPES)

	def __contains__(self, item):
		if isinstance(item, DmgSaveTypes):
			mbc = item.GetMbc()
		else:
			mbc = item
		return self._FindByMbc(mbc) is not None

	@classmethod
	def GetCLINames(cls, include_auto=True, include_batteryless=True):
		# Names are listed in RAM_TYPES order so the CLI menu is stable.
		out = ["auto"] if include_auto else []
		for entry in cls.RAM_TYPES:
			mbc = entry[0]
			if mbc == 0x205 and not include_batteryless:
				continue
			name = cls.CLI_NAMES.get(mbc)
			if name is not None:
				out.append(name)
		return out

	@classmethod
	def GetMbcFromCLIName(cls, name):
		if name == "auto":
			return None
		for mbc, n in cls.CLI_NAMES.items():
			if n == name:
				return mbc
		return None
